# Copyright 2020-2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs, Keegan Smith

from __future__ import annotations

import logging
import os
from typing import Dict, List, Tuple, Union

import pendulum
from airflow.exceptions import AirflowException, AirflowSkipException
from airflow.hooks.base import BaseHook
from google.cloud.bigquery import TimePartitioningType, SourceFormat, WriteDisposition
from googleapiclient.discovery import Resource, build
from oauth2client.service_account import ServiceAccountCredentials

from oaebu_workflows.oaebu_partners import OaebuPartner, partner_from_str
from observatory.api.client.model.dataset_release import DatasetRelease
from observatory.platform.api import make_observatory_api
from observatory.platform.files import save_jsonl_gz
from observatory.platform.airflow import AirflowConns
from observatory.platform.files import add_partition_date
from observatory.platform.gcs import gcs_upload_files, gcs_blob_uri, gcs_blob_name_from_path
from observatory.platform.bigquery import bq_load_table, bq_table_id, bq_create_dataset
from observatory.platform.observatory_config import CloudWorkspace
from observatory.platform.workflows.workflow import (
    Workflow,
    PartitionRelease,
    cleanup,
    set_task_state,
    check_workflow_inputs,
)


class GoogleAnalytics3Release(PartitionRelease):
    def __init__(
        self,
        dag_id: str,
        run_id: str,
        data_interval_start: pendulum.DateTime,
        data_interval_end: pendulum.DateTime,
        partition_date: pendulum.DateTime,
    ):
        """Construct a GoogleAnalytics3Release.

        :param dag_id: The ID of the DAG
        :param run_id: The Airflow run ID
        :param data_interval_start: The start date of the DAG the start date of the download period.
        :param data_interval_end: end date of the download period, also used as release date for BigQuery table and file paths
        """
        super().__init__(dag_id=dag_id, run_id=run_id, partition_date=partition_date)
        self.data_interval_start = data_interval_start
        self.data_interval_end = data_interval_end
        self.transform_path = os.path.join(self.transform_folder, f"{partition_date.format('YYYY_MM_DD')}.json.gz")


class GoogleAnalytics3Telescope(Workflow):
    """Google Analytics Telescope."""

    ANU_ORG_NAME = "ANU Press"

    def __init__(
        self,
        dag_id: str,
        organisation_name: str,
        cloud_workspace: CloudWorkspace,
        view_id: str,
        pagepath_regex: str,
        data_partner: Union[str, OaebuPartner] = "google_analytics3",
        bq_dataset_description: str = "Data from Google sources",
        bq_table_description: str = None,
        api_dataset_id: str = "google_analytics",
        oaebu_service_account_conn_id: str = "oaebu_service_account",
        observatory_api_conn_id: str = AirflowConns.OBSERVATORY_API,
        catchup: bool = True,
        start_date: pendulum.DateTime = pendulum.datetime(2018, 1, 1),
        schedule: str = "@monthly",
    ):
        """Construct a GoogleAnalytics3Telescope instance.
        :param dag_id: The ID of the DAG
        :param organisation_name: The organisation name as per Google Analytics
        :param cloud_workspace: The CloudWorkspace object for this DAG
        :param view_id: The Google Analytics view ID
        :param pagepath_regex: The pagepath regex
        :param data_partner: The name of the data partner
        :param bq_dataset_description: Description for the BigQuery dataset
        :param bq_table_description: Description for the biguery table
        :param api_dataset_id: The ID to store the dataset release in the API
        :param oaebu_service_account_conn_id: Airflow connection ID for the OAEBU service account
        :param observatory_api_conn_id: Airflow connection ID for the overvatory API
        :param catchup: Whether to catchup the DAG or not
        :param start_date: The start date of the DAG
        :param schedule: The schedule interval of the DAG
        """
        super().__init__(
            dag_id,
            start_date,
            schedule,
            catchup=catchup,
            airflow_conns=[oaebu_service_account_conn_id, observatory_api_conn_id],
            tags=["oaebu"],
        )

        self.dag_id = dag_id
        self.organisation_name = organisation_name
        self.cloud_workspace = cloud_workspace
        self.view_id = view_id
        self.pagepath_regex = pagepath_regex
        self.data_partner = partner_from_str(data_partner)
        self.bq_dataset_description = bq_dataset_description
        self.bq_table_description = bq_table_description
        self.api_dataset_id = api_dataset_id
        self.oaebu_service_account_conn_id = oaebu_service_account_conn_id
        self.observatory_api_conn_id = observatory_api_conn_id

        check_workflow_inputs(self)

        self.add_setup_task(self.check_dependencies)
        self.add_task(self.download_transform)
        self.add_task(self.upload_transformed)
        self.add_task(self.bq_load)
        self.add_task(self.add_new_dataset_releases)
        self.add_task(self.cleanup)

    def make_release(self, **kwargs) -> List[GoogleAnalytics3Release]:
        """Make release instances. The release is passed as an argument to the function (TelescopeFunction) that is
        called in 'task_callable'.

        :param kwargs: the context passed from the PythonOperator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for the keyword arguments that can be passed
        :return: A list of grid release instances
        """
        # Get start and end date (data_interval_end = release_date)
        data_interval_start = kwargs["data_interval_start"].start_of("month")
        data_interval_end = kwargs["data_interval_end"].start_of("month")
        partition_date = data_interval_start.end_of("month")

        logging.info(
            f"Start date: {data_interval_start}, end date:{data_interval_end}, parition_date: {partition_date}"
        )
        releases = [
            GoogleAnalytics3Release(
                dag_id=self.dag_id,
                run_id=kwargs["run_id"],
                data_interval_start=data_interval_start,
                data_interval_end=data_interval_end,
                partition_date=partition_date,
            )
        ]
        return releases

    def check_dependencies(self, **kwargs) -> bool:
        """Check dependencies of DAG. Add to parent method to additionally check for a view id and pagepath regex

        :param kwargs: the context passed from the Airflow Operator.
        :return: True if dependencies are valid.
        """
        super().check_dependencies()

        if self.view_id is None or self.pagepath_regex is None:
            expected_extra = {"view_id": "the_view_id", "pagepath_regex": r"pagepath_regex"}
            raise AirflowException(
                f"View ID and/or pagepath regex is not set in 'extra' of telescope, extra example: " f"{expected_extra}"
            )
        return True

    def download_transform(self, releases: List[GoogleAnalytics3Release], **kwargs) -> None:
        """Task to download and transform the google analytics release for a given month.

        :param releases: a list with one google analytics release.
        """
        data_found = False
        for release in releases:
            service = initialize_analyticsreporting(self.oaebu_service_account_conn_id)
            results = get_reports(
                service,
                self.organisation_name,
                self.view_id,
                self.pagepath_regex,
                release.data_interval_start,
                release.data_interval_end.subtract(
                    days=1
                ),  # Subtract 1 day because GA uses inclusive dates, Airlfow data intervals are not
            )
            results = add_partition_date(
                results, release.partition_date, TimePartitioningType.MONTH, partition_field="release_date"
            )
            if results:
                save_jsonl_gz(release.transform_path, results)
                data_found = True
            else:
                if (pendulum.today("UTC") - self.data_interval_end).in_months() >= 26:
                    logging.info(
                        "No data available. Google Analytics data is only available for 26 months, see "
                        "https://support.google.com/analytics/answer/7667196?hl=en for more info"
                    )

        if not data_found:
            raise AirflowSkipException("No Google Analytics data available to download.")

    def upload_transformed(self, releases: List[GoogleAnalytics3Release], **kwargs) -> None:
        """Uploads the transformed file to GCS"""
        for release in releases:
            state = gcs_upload_files(
                bucket_name=self.cloud_workspace.transform_bucket, file_paths=[release.transform_path]
            )
            set_task_state(state, kwargs["ti"].task_id, release=release)

    def bq_load(self, releases: List[GoogleAnalytics3Release], **kwargs) -> None:
        """Loads the data into BigQuery"""
        bq_create_dataset(
            project_id=self.cloud_workspace.project_id,
            dataset_id=self.data_partner.bq_dataset_id,
            location=self.cloud_workspace.data_location,
            description=self.bq_dataset_description,
        )
        for release in releases:
            uri = gcs_blob_uri(self.cloud_workspace.transform_bucket, gcs_blob_name_from_path(release.transform_path))
            table_id = bq_table_id(
                self.cloud_workspace.project_id, self.data_partner.bq_dataset_id, self.data_partner.bq_table_name
            )
            state = bq_load_table(
                uri=uri,
                table_id=table_id,
                schema_file_path=self.data_partner.schema_path,
                source_format=SourceFormat.NEWLINE_DELIMITED_JSON,
                partition_type=TimePartitioningType.MONTH,
                partition=True,
                partition_field="release_date",
                write_disposition=WriteDisposition.WRITE_APPEND,
                table_description=self.bq_table_description,
                ignore_unknown_values=True,
            )
            set_task_state(state, kwargs["ti"].task_id, release=release)

    def add_new_dataset_releases(self, releases: List[GoogleAnalytics3Release], **kwargs) -> None:
        """Adds release information to API."""
        api = make_observatory_api(observatory_api_conn_id=self.observatory_api_conn_id)
        for release in releases:
            dataset_release = DatasetRelease(
                dag_id=self.dag_id,
                dataset_id=self.api_dataset_id,
                dag_run_id=release.run_id,
                data_interval_start=release.data_interval_start,
                data_interval_end=release.data_interval_end,
                partition_date=release.partition_date,
            )
            api.post_dataset_release(dataset_release)

    def cleanup(self, releases: List[GoogleAnalytics3Release], **kwargs) -> None:
        """Delete all files, folders and XComs associated with this release."""
        for release in releases:
            cleanup(
                dag_id=self.dag_id, execution_date=kwargs["execution_date"], workflow_folder=release.workflow_folder
            )


def initialize_analyticsreporting(oaebu_service_account_conn_id: str) -> Resource:
    """Initializes an Analytics Reporting API V4 service object.

    :return: An authorized Analytics Reporting API V4 service object.
    """
    oaebu_account_conn = BaseHook.get_connection(oaebu_service_account_conn_id)

    scopes = ["https://www.googleapis.com/auth/analytics.readonly"]
    creds = ServiceAccountCredentials.from_json_keyfile_dict(oaebu_account_conn.extra_dejson, scopes=scopes)

    # Build the service object.
    service = build("analyticsreporting", "v4", credentials=creds, cache_discovery=False)

    return service


def list_all_books(
    service: Resource,
    view_id: str,
    pagepath_regex: str,
    data_interval_start: pendulum.DateTime,
    data_interval_end: pendulum.DateTime,
    organisation_name: str,
    metrics: list,
) -> Tuple[List[dict], list]:
    """List all available books by getting all pagepaths of a view id in a given period.
    Note: Google API will not return a result for any entry in which all supplied metrics are zero.
    However, it will return 'some' results if you supply no metrics, contrary to the documentation.
    Date ranges are inclusive.

    :param service: The Google Analytics Reporting service object.
    :param view_id: The view id.
    :param pagepath_regex: The regex expression for the pagepath of a book.
    :param data_interval_start: The start date of the DAG Start date of analytics period
    :param data_interval_end: End date of analytics period
    :param organisation_name: The organisation name.
    :param: metrics: The metrics to return return with the book results
    :return: A list with dictionaries, one for each book entry (the dict contains the pagepath, title and average time
    on page) and a list of all pagepaths.
    """
    # Get pagepath, pagetitle and average time on page for each path
    body = {
        "reportRequests": [
            {
                "viewId": view_id,
                "pageSize": 10000,
                "dateRanges": [
                    {
                        "startDate": data_interval_start.strftime("%Y-%m-%d"),
                        "endDate": data_interval_end.strftime("%Y-%m-%d"),
                    }
                ],
                "metrics": metrics,
                "dimensions": [{"name": "ga:pagepath"}, {"name": "ga:pageTitle"}],
                "dimensionFilterClauses": [
                    {
                        "operator": "AND",
                        "filters": [
                            {"dimensionName": "ga:pagepath", "operator": "REGEXP", "expressions": [pagepath_regex]}
                        ],
                    }
                ],
            }
        ]
    }

    # add all 6 custom dimensions for anu press
    if organisation_name == GoogleAnalytics3Telescope.ANU_ORG_NAME:
        for i in range(1, 7):
            body["reportRequests"][0]["dimensions"].append({"name": f"ga:dimension{str(i)}"})

    reports = service.reports().batchGet(body=body).execute()
    all_book_entries = reports["reports"][0]["data"].get("rows")
    next_page_token = reports["reports"][0].get("nextPageToken")

    while next_page_token:
        body["reportRequests"][0]["pageToken"] = next_page_token
        reports = service.reports().batchGet(body=body).execute()
        book_entries = reports["reports"][0]["data"].get("rows")
        next_page_token = reports["reports"][0].get("nextPageToken")
        all_book_entries += book_entries

    # create list with just pagepaths
    if all_book_entries:
        pagepaths = [path["dimensions"][0] for path in all_book_entries]
    else:
        pagepaths = []

    return all_book_entries, pagepaths


def create_book_result_dicts(
    book_entries: List[dict],
    data_interval_start: pendulum.DateTime,
    data_interval_end: pendulum.DateTime,
    organisation_name: str,
) -> Dict[dict]:
    """Create a dictionary to store results for a single book. Pagepath, title and avg time on page are already given.
    The other metrics will be added to the dictionary later.

    :param book_entries: List with dictionaries of book entries.
    :param data_interval_start: The start date of the DAG Start date of analytics period.
    :param data_interval_end: End date of analytics period.
    :param organisation_name: The organisation name.
    :return: Dict to store results
    """
    book_results = {}
    for entry in book_entries:
        pagepath = entry["dimensions"][0]
        pagetitle = entry["dimensions"][1]
        average_time = float(entry["metrics"][0]["values"][-1])
        book_result = {
            "url": pagepath,
            "title": pagetitle,
            "start_date": data_interval_start.strftime("%Y-%m-%d"),
            "end_date": data_interval_end.strftime("%Y-%m-%d"),
            "average_time": average_time,
            "unique_views": {"country": {}, "referrer": {}, "social_network": {}},
            "page_views": {"country": {}, "referrer": {}, "social_network": {}},
            "sessions": {"country": {}, "source": {}},
        }
        # add custom dimension data for ANU Press
        if organisation_name == GoogleAnalytics3Telescope.ANU_ORG_NAME:
            # matches dimension order in 'list_all_books'
            custom_dimensions = {
                "publication_id": entry["dimensions"][2],
                "publication_type": entry["dimensions"][3],
                "publication_imprint": entry["dimensions"][4],
                "publication_group": entry["dimensions"][5],
                "publication_whole_or_part": entry["dimensions"][6],
                "publication_format": entry["dimensions"][7],
            }
            book_result = dict(book_result, **custom_dimensions)
        book_results[pagepath] = book_result

    return book_results


def get_dimension_data(
    service: Resource,
    view_id: str,
    data_interval_start: pendulum.DateTime,
    data_interval_end: pendulum.DateTime,
    metrics: list,
    dimension: dict,
    pagepaths: list,
) -> list:
    """Get reports data from the Google Analytics Reporting service for a single dimension and multiple metrics.
    The results are filtered by pagepaths of interest and ordered by pagepath as well.

    :param service: The Google Analytics Reporting service.
    :param view_id: The view id.
    :param data_interval_start: The start date of the DAG The start date of the analytics period.
    :param data_interval_end: The end date of the analytics period.
    :param metrics: List with dictionaries of metric.
    :param dimension: The dimension.
    :param pagepaths: List with pagepaths to filter and sort on.
    :return: List with reports data for dimension and metrics.
    """
    body = {
        "reportRequests": [
            {
                "viewId": view_id,
                "pageSize": 10000,
                "dateRanges": [
                    {
                        "startDate": data_interval_start.strftime("%Y-%m-%d"),
                        "endDate": data_interval_end.strftime("%Y-%m-%d"),
                    }
                ],
                "metrics": metrics,
                "dimensions": [{"name": "ga:pagePath"}, dimension],
                "dimensionFilterClauses": [
                    {"filters": [{"dimensionName": "ga:pagePath", "operator": "IN_LIST", "expressions": pagepaths}]}
                ],
                "orderBys": [{"fieldName": "ga:pagepath"}],
            }
        ]
    }
    reports = service.reports().batchGet(body=body).execute()
    all_dimension_data = reports["reports"][0]["data"].get("rows")
    next_page_token = reports["reports"][0].get("nextPageToken")

    while next_page_token:
        body["reportRequests"][0]["pageToken"] = next_page_token
        reports = service.reports().batchGet(body=body).execute()
        dimension_data = reports["reports"][0]["data"].get("rows")
        next_page_token = reports["reports"][0].get("nextPageToken")
        all_dimension_data += dimension_data

    return all_dimension_data


def add_to_book_result_dict(
    book_results: dict, dimension: dict, pagepath: str, unique_views: dict, page_views: dict, sessions: dict
):
    """Add the 'unique_views', 'page_views' and 'sessions' results to the book results dict if these metrics are of interest for the
    current dimension.

    :param book_results: A dictionary with all book results.
    :param dimension: Current dimension for which 'unique_views' and 'sessions' data is given.
    :param pagepath: Pagepath of the book.
    :param unique_views: Number of unique views for the pagepath&dimension
    :param page_views: Number of page views for the pagepath&dimension
    :param sessions: Number of sessions for the pagepath&dimension
    :return: None
    """
    # map the dimension name to the field name in BigQuery. The ga:dimensionX are obtained from custom ANU press
    # dimensions
    mapping = {
        "ga:country": "country",
        "ga:fullReferrer": "referrer",
        "ga:socialNetwork": "social_network",
        "ga:source": "source",
    }
    column_name = mapping[dimension["name"]]
    if column_name in ["country", "referrer", "social_network"]:
        book_results[pagepath]["unique_views"][column_name] = unique_views
        book_results[pagepath]["page_views"][column_name] = page_views
    if column_name in ["country", "source"]:
        book_results[pagepath]["sessions"][column_name] = sessions


def get_reports(
    service: Resource,
    organisation_name: str,
    view_id: str,
    pagepath_regex: str,
    data_interval_start: pendulum.DateTime,
    data_interval_end: pendulum.DateTime,
) -> list:
    """Get reports data from the Google Analytics Reporting API.

    :param service: The Google Analytics Reporting service.
    :param organisation_name: Name of the organisation.
    :param view_id: The view id.
    :param pagepath_regex: The regex expression for the pagepath of a book.
    :param data_interval_start: The start date of the DAG Start date of analytics period
    :param data_interval_end: End date of analytics period
    :return: List with google analytics data for each book
    """

    metric_names = ["uniquePageviews", "Pageviews", "sessions", "avgTimeOnPage"]
    metrics = [{"expression": f"ga:{metric}"} for metric in metric_names]

    # list all books
    book_entries, pagepaths = list_all_books(
        service, view_id, pagepath_regex, data_interval_start, data_interval_end, organisation_name, metrics
    )
    # if no books in period return empty list and raise airflow skip exception
    if not book_entries:
        return []
    # create dict with dict for each book to store results
    book_results = create_book_result_dicts(book_entries, data_interval_start, data_interval_end, organisation_name)

    dimension_names = ["country", "fullReferrer", "socialNetwork", "source"]
    dimensions = [{"name": f"ga:{dimension}"} for dimension in dimension_names]

    # get data per dimension
    for dimension in dimensions:
        dimension_data = get_dimension_data(
            service, view_id, data_interval_start, data_interval_end, metrics, dimension, pagepaths
        )

        prev_pagepath = None
        unique_views = {}
        page_views = {}
        sessions = {}
        # entry is combination of book pagepath & dimension
        for entry in dimension_data:
            pagepath = entry["dimensions"][0]
            dimension_value = entry["dimensions"][1]  # e.g. 'Australia' for 'country' dimension

            if prev_pagepath and pagepath != prev_pagepath:
                add_to_book_result_dict(book_results, dimension, prev_pagepath, unique_views, page_views, sessions)

                unique_views = {}
                page_views = {}
                sessions = {}

            # add values if they are not 0
            # ["values"][n] maps to the nth value of metric_names
            unique_views_metric = int(entry["metrics"][0]["values"][0])
            page_views_metric = int(entry["metrics"][0]["values"][1])
            sessions_metric = int(entry["metrics"][0]["values"][2])
            if unique_views_metric > 0:
                unique_views[dimension_value] = unique_views_metric
            if page_views_metric > 0:
                page_views[dimension_value] = page_views_metric
            if sessions_metric > 0:
                sessions[dimension_value] = sessions_metric

            prev_pagepath = pagepath
        else:
            add_to_book_result_dict(book_results, dimension, prev_pagepath, unique_views, page_views, sessions)

    # transform nested dict to list of dicts
    for book, result in book_results.items():
        for field, value in result.items():
            # field is 'unique_views' or 'sessions'
            if isinstance(value, dict):
                # nested_field is 'country', 'referrer' or 'social_network'
                for nested_field, nested_value in value.items():
                    values = []
                    # k is e.g. 'Australia', v is e.g. 1
                    for k, v in nested_value.items():
                        values.append({"name": k, "value": v})
                    book_results[book][field][nested_field] = values

    # convert dict to list of results
    book_results = [book_results[k] for k in book_results]

    return book_results



# Copyright 2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs, Keegan Smith


import logging
import os
from typing import List, Union
from concurrent.futures import ThreadPoolExecutor, as_completed

import pendulum
from airflow.hooks.base import BaseHook
from google.cloud.bigquery import SourceFormat, TimePartitioningType, WriteDisposition
from google.oauth2 import service_account
from apiclient import discovery

from oaebu_workflows.oaebu_partners import OaebuPartner, partner_from_str
from observatory.api.client.model.dataset_release import DatasetRelease
from observatory.platform.api import make_observatory_api
from observatory.platform.airflow import AirflowConns
from observatory.platform.files import save_jsonl_gz, load_jsonl
from observatory.platform.gcs import gcs_blob_uri, gcs_upload_files, gcs_blob_name_from_path
from observatory.platform.bigquery import bq_load_table, bq_table_id, bq_create_dataset
from observatory.platform.utils.url_utils import retry_get_url
from observatory.platform.observatory_config import CloudWorkspace
from observatory.platform.files import add_partition_date
from observatory.platform.workflows.workflow import (
    Workflow,
    PartitionRelease,
    cleanup,
    set_task_state,
    check_workflow_inputs,
)


class UclDiscoveryRelease(PartitionRelease):
    def __init__(
        self,
        dag_id: str,
        run_id: str,
        data_interval_start: pendulum.DateTime,
        data_interval_end: pendulum.DateTime,
        partition_date: pendulum.DateTime,
    ):
        """Construct a UclDiscoveryRelease instance.

        :param dag_id: The ID of the DAG
        :param run_id: The Airflow run ID.
        :param data_interval_start: The start of the data interval.
        :param data_interval_end: The end of the data interval.
        :param partition_date: The partition date for this release.
        """
        super().__init__(dag_id=dag_id, run_id=run_id, partition_date=partition_date)
        self.data_interval_start = data_interval_start
        self.data_interval_end = data_interval_end

        self.download_country_path = os.path.join(self.download_folder, "ucl_discovery_country.jsonl.gz")
        self.download_totals_path = os.path.join(self.download_folder, "ucl_discovery_totals.jsonl.gz")
        self.transform_path = os.path.join(self.transform_folder, "ucl_discovery.jsonl.gz")


class UclDiscoveryTelescope(Workflow):
    """The UCL Discovery telescope."""

    def __init__(
        self,
        dag_id: str,
        cloud_workspace: CloudWorkspace,
        sheet_id: str,
        data_partner: Union[str, OaebuPartner] = "ucl_discovery",
        bq_dataset_description: str = "UCL Discovery dataset",
        bq_table_description: str = "UCL Discovery table",
        api_dataset_id: str = "ucl",
        observatory_api_conn_id: str = AirflowConns.OBSERVATORY_API,
        oaebu_service_account_conn_id: str = "oaebu_service_account",
        max_threads: int = os.cpu_count() * 2,
        schedule: str = "0 0 4 * *",  # run on the 4th of every month
        start_date: pendulum.DateTime = pendulum.datetime(2015, 6, 1),
        catchup: bool = True,
        max_active_runs: int = 10,
    ):
        """Construct a UclDiscoveryTelescope instance.

        :param dag_id: The ID of the DAG
        :param cloud_workspace: The CloudWorkspace object for this DAG
        :param sheet_id:  The ID of the google sheet match eprint ID to ISBN13
        :param data_partner: The name of the data partner
        :param bq_dataset_description: Description for the BigQuery dataset
        :param bq_table_description: Description for the biguery table
        :param api_dataset_id: The ID to store the dataset release in the API
        :param observatory_api_conn_id: Airflow connection ID for the overvatory API
        :param oaebu_service_account_conn_id: Airflow connection ID for the oaebu service account
        :param max_threads: The maximum number threads to utilise for parallel processes
        :param schedule: The schedule interval of the DAG
        :param start_date: The start date of the DAG
        :param catchup: Whether to catchup the DAG or not
        :param max_active_runs: The maximum number of concurrent DAG runs
        """
        super().__init__(
            dag_id,
            start_date,
            schedule,
            catchup=catchup,
            max_active_runs=max_active_runs,
            airflow_conns=[observatory_api_conn_id, oaebu_service_account_conn_id],
            tags=["oaebu"],
        )

        self.dag_id = dag_id
        self.cloud_workspace = cloud_workspace
        self.sheet_id = sheet_id
        self.data_partner = partner_from_str(data_partner)
        self.bq_dataset_description = bq_dataset_description
        self.bq_table_description = bq_table_description
        self.api_dataset_id = api_dataset_id
        self.oaebu_service_account_conn_id = oaebu_service_account_conn_id
        self.max_threads = max_threads
        self.observatory_api_conn_id = observatory_api_conn_id

        check_workflow_inputs(self)

        self.add_setup_task(self.check_dependencies)
        self.add_task(self.download)
        self.add_task(self.upload_downloaded)
        self.add_task(self.transform)
        self.add_task(self.upload_transformed)
        self.add_task(self.bq_load)
        self.add_task(self.add_new_dataset_releases)
        self.add_task(self.cleanup)

    def make_release(self, **kwargs) -> List[UclDiscoveryRelease]:
        """Make release instances. The release is passed as an argument to the function (TelescopeFunction) that is
        called in 'task_callable'. There will only be 1 release, but it is passed on as a list so the
        SnapshotTelescope template methods can be used.

        :param kwargs: the context passed from the PythonOperator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for the keyword arguments that can be passed
        :return: A list with one ucldiscovery release instance.
        """
        data_interval_start = kwargs["data_interval_start"].start_of("month")
        data_interval_end = kwargs["data_interval_end"].start_of("month")
        partition_date = data_interval_start.end_of("month")
        run_id = kwargs["run_id"]

        logging.info(
            f"Interval Start: {data_interval_start}, Interval End:{data_interval_end}, Partition date: {partition_date}, Run ID: {run_id}"
        )
        return UclDiscoveryRelease(
            self.dag_id,
            kwargs["run_id"],
            data_interval_start=data_interval_start,
            data_interval_end=data_interval_end,
            partition_date=partition_date,
        )

    def download(self, release: UclDiscoveryRelease, **kwargs):
        """Fownload the ucl discovery data for a given release.
        :param releases: The UCL discovery release.
        """
        mappings = get_isbn_eprint_mappings(self.sheet_id, self.oaebu_service_account_conn_id, release.partition_date)

        with ThreadPoolExecutor(self.max_threads) as executor:
            futures = []
            for eprint_id in mappings.keys():
                future = executor.submit(
                    download_discovery_stats, eprint_id, release.data_interval_start, release.partition_date
                )
                futures.append(future)
            totals = []
            country = []
            for future in as_completed(futures):
                c, t = future.result()
                country.append(c)
                totals.append(t)

        logging.info(f"Saving totals data to file: {release.download_totals_path}")
        save_jsonl_gz(release.download_totals_path, totals)
        logging.info(f"Saving country data to file: {release.download_country_path}")
        save_jsonl_gz(release.download_country_path, country)

    def upload_downloaded(self, release: UclDiscoveryRelease, **kwargs):
        """Uploads the downloaded files to GCS"""
        success = gcs_upload_files(
            bucket_name=self.cloud_workspace.download_bucket,
            file_paths=[release.download_country_path, release.download_totals_path],
        )
        set_task_state(success, kwargs["ti"].task_id, release=release)

    def transform(self, release: UclDiscoveryRelease, **kwargs):
        """Transform the ucl discovery data for a given release."""
        mappings = get_isbn_eprint_mappings(self.sheet_id, self.oaebu_service_account_conn_id, release.partition_date)

        # Load the records and sort them by eprint id
        country_records = load_jsonl(release.download_country_path)
        totals_records = load_jsonl(release.download_totals_path)
        country_records = sorted(country_records, key=lambda x: x["set"]["value"])  # ["set"]["value"] = eprint_id
        totals_records = sorted(totals_records, key=lambda x: x["set"]["value"])
        assert len(country_records) == len(totals_records), f"{len(country_records)} != {len(totals_records)}"

        with ThreadPoolExecutor(self.max_threads) as executor:
            futures = []
            for country_record, totals_record in zip(country_records, totals_records):
                isbn = mappings[country_record["set"]["value"]]["ISBN13"]
                title = mappings[country_record["set"]["value"]]["title"]
                future = executor.submit(transform_discovery_stats, country_record, totals_record, isbn, title)
                futures.append(future)
            results = []
            for future in as_completed(futures):
                results.append(future.result())

        # Add the release date to the data as a parition field
        results = add_partition_date(
            results, release.partition_date, TimePartitioningType.MONTH, partition_field="release_date"
        )
        print(results)
        save_jsonl_gz(release.transform_path, results)

    def upload_transformed(self, release: UclDiscoveryRelease, **kwargs):
        """Uploads the transformed file to GCS"""
        success = gcs_upload_files(
            bucket_name=self.cloud_workspace.transform_bucket, file_paths=[release.transform_path]
        )
        set_task_state(success, kwargs["ti"].task_id, release=release)

    def bq_load(self, release: UclDiscoveryRelease, **kwargs) -> None:
        """Loads the transformed data into BigQuery"""
        bq_create_dataset(
            project_id=self.cloud_workspace.project_id,
            dataset_id=self.data_partner.bq_dataset_id,
            location=self.cloud_workspace.data_location,
            description=self.bq_dataset_description,
        )

        uri = gcs_blob_uri(self.cloud_workspace.transform_bucket, gcs_blob_name_from_path(release.transform_path))
        table_id = bq_table_id(
            self.cloud_workspace.project_id, self.data_partner.bq_dataset_id, self.data_partner.bq_table_name
        )
        state = bq_load_table(
            uri=uri,
            table_id=table_id,
            schema_file_path=self.data_partner.schema_path,
            source_format=SourceFormat.NEWLINE_DELIMITED_JSON,
            partition_type=TimePartitioningType.MONTH,
            partition=True,
            partition_field="release_date",
            write_disposition=WriteDisposition.WRITE_APPEND,
            table_description=self.bq_table_description,
            ignore_unknown_values=True,
        )
        set_task_state(state, kwargs["ti"].task_id, release=release)

    def add_new_dataset_releases(self, release: UclDiscoveryRelease, **kwargs) -> None:
        """Adds release information to API."""
        api = make_observatory_api(observatory_api_conn_id=self.observatory_api_conn_id)
        dataset_release = DatasetRelease(
            dag_id=self.dag_id,
            dataset_id=self.api_dataset_id,
            dag_run_id=release.run_id,
            data_interval_start=kwargs["data_interval_start"],
            data_interval_end=kwargs["data_interval_end"],
            partition_date=release.partition_date,
        )
        api.post_dataset_release(dataset_release)

    def cleanup(self, release: UclDiscoveryRelease, **kwargs) -> None:
        """Delete all files, folders and XComs associated with this release."""
        cleanup(dag_id=self.dag_id, execution_date=kwargs["execution_date"], workflow_folder=release.workflow_folder)


def get_isbn_eprint_mappings(sheet_id: str, service_account_conn_id: str, cutoff_date: pendulum.DateTime) -> dict:
    """Get the eprint id to isbn mapping from the google sheet

    :param sheet_id: The ID of the google sheet.
    :param credentials: The credentials object to authenticate with.
    :param cutoff_date: The cutoff date. If an item is published after this date, it will be skipped.
    """
    scopes = [
        "https://www.googleapis.com/auth/drive",
        "https://www.googleapis.com/auth/drive.file",
        "https://www.googleapis.com/auth/spreadsheets",
    ]
    service_account_conn = BaseHook.get_connection(service_account_conn_id)
    credentials = service_account.Credentials.from_service_account_info(
        service_account_conn.extra_dejson, scopes=scopes
    )
    service = discovery.build("sheets", "v4", credentials=credentials)
    result = service.spreadsheets().values().get(spreadsheetId=sheet_id, range="isbn_mapping").execute()
    sheet_contents = result.get("values")
    if not sheet_contents:
        raise ValueError(f"No content found for sheet with ID {sheet_id}")

    items = []
    header = sheet_contents[0]
    if not all(heading in header for heading in ["ISBN13", "discovery_eprintid", "date", "title_list_title"]):
        raise ValueError(f"Invalid header found for sheet: {header}")
    for row in sheet_contents[1:]:
        items.append(dict(zip(header, row)))

    mappings = {}
    for item in items:
        eprint_id = item.get("discovery_eprintid")
        isbn = item.get("ISBN13")
        title = item.get("title_list_title")
        if not eprint_id or not isbn:
            logging.warn(f"Item with missing information will be skipped: {item}")
            continue
        if pendulum.parse(item["date"]) > cutoff_date:
            logging.info(f"Item released after cutoff date will be skipped: {item}")
            continue
        mappings[eprint_id] = {"ISBN13": isbn, "title": title}

    return mappings


def download_discovery_stats(eprint_id: str, start_date: pendulum.DateTime, end_date: pendulum.DateTime):
    """Downloads the discovery stats for a given eprint ID within a specified date range.

    :param eprint_id: The eprint ID of the item to get the stats for.
    :param start_date: The start date of the date range.
    :param end_date: The end date of the date range.
    :return: A tuple containing the country statistics and the total downloads statistics.
    """
    countries_url = (
        "https://discovery.ucl.ac.uk/cgi/stats/get"
        f"?from={start_date.format('YYYYMMDD')}&to={end_date.format('YYYYMMDD')}"
        f"&irs2report=eprint&set_name=eprint&set_value={eprint_id}&datatype=countries&top=countries"
        "&view=Table&limit=all&export=JSON"
    )
    totals_url = (
        "https://discovery.ucl.ac.uk/cgi/stats/get"
        f"?from={start_date.format('YYYYMMDD')}&to={end_date.format('YYYYMMDD')}"
        f"&irs2report=eprint&set_name=eprint&set_value={eprint_id}&datatype=downloads&graph_type=column"
        "&view=Google%3A%3AGraph&date_resolution=month&title=Download+activity+-+last+12+months&export=JSON"
    )
    response = retry_get_url(countries_url)
    country = response.json()
    response = retry_get_url(totals_url)
    totals = response.json()

    # Perform some checks on the returned data
    timescale = (start_date.format("YYYYMMDD"), end_date.format("YYYYMMDD"))
    country_timescale = (country["timescale"]["from"], country["timescale"]["to"])
    totals_timescale = (totals["timescale"]["from"], totals["timescale"]["to"])
    if country_timescale != timescale or totals_timescale != timescale:
        raise ValueError(
            f"Invalid timescale value(s): country: {country['timescale']} | totals: {totals['timescale']} != {timescale}"
        )
    if country["set"]["value"] != eprint_id or totals["set"]["value"] != eprint_id:
        raise ValueError(
            f"Invalid eprint ID values downloaded: {totals['set']['value']} | {country['set']['value']} != {eprint_id}"
        )

    return country, totals


def transform_discovery_stats(country_record: dict, totals_record: dict, isbn: str, title: str) -> dict:
    """Transforms the discovery stats for a single set of records

    :param country_record: The country record
    :param totals_record: The totals record
    :param isbn: The isbn that matches the eprint id
    :return: The transformed stats
    """
    # Sanity check the records
    country_eprint_id = country_record["set"]["value"]
    totals_eprint_id = totals_record["set"]["value"]
    if country_eprint_id != totals_eprint_id:
        raise ValueError(f"Country and totals eprint ID do not match: {country_eprint_id} != {totals_eprint_id}")

    country_timescale = country_record["timescale"]
    totals_timescale = totals_record["timescale"]
    if country_timescale != totals_timescale:
        raise ValueError(f"Timescales do not match: {country_timescale} != {totals_timescale}")

    # If there are no downloads for the time period, there is no "records" field in country stats
    country_records = country_record.get("records", [])

    transformed = {
        "ISBN": isbn,
        "title": title,
        "eprint_id": totals_record["set"]["value"],
        "timescale": totals_record["timescale"],
        "origin": totals_record["origin"],
        "total_downloads": totals_record["records"][0]["count"],
        "country": country_records,
    }
    return transformed



# Copyright 2020-2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose, Keegan Smith

import os
from typing import Optional

from observatory.platform.config import module_file_path


def test_fixtures_folder(workflow_module: Optional[str] = None) -> str:
    """Get the path to the OAeBU Workflows test data directory.

    :param workflow_module: Optional, name of the workflow. Only to be included if the schema for the workflow is in
    the directory oaebu_workflows.{workflow_name}.tests.fixtures
    :return: the test fixtures directory.
    """

    if workflow_module is not None:
        return construct_module_path("oaebu_workflows", workflow_module, "tests", "fixtures")

    return construct_module_path("oaebu_workflows", "tests", "fixtures")


def schema_folder(workflow_module: Optional[str] = None) -> str:
    """Return the path to the schema template folder.

    :param workflow_module: Optional, name of the workflow. Only to be included if the schema for the workflow is in
    the directory oaebu_workflows.{workflow_module}.schema
    :return: the path.
    """

    if workflow_module is not None:
        return construct_module_path("oaebu_workflows", workflow_module, "schema")

    return construct_module_path("oaebu_workflows", "schema")


def sql_folder(workflow_module: Optional[str] = None) -> str:
    """Return the path to the workflow SQL template folder.

    :param workflow_module: Optional, name of the workflow. Only to be included if the sql for the workflow is in
    the directory oaebu_workflows.{workflow_module}.sql
    :return: the path.
    """

    if workflow_module is not None:
        return construct_module_path("oaebu_workflows", workflow_module, "sql")

    return construct_module_path("oaebu_workflows", "sql")


def construct_module_path(*parts: str) -> str:
    """Constructs the full module path given parts of a path."""

    module_path = ".".join(list(parts))
    file_path = module_file_path(module_path)
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"construct_module_path: directory {file_path} does not exist!")

    return file_path



# Copyright 2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Author: Keegan Smith

import unittest
from unittest.mock import patch

from oaebu_workflows.oaebu_partners import OaebuPartner, partner_from_str, OAEBU_DATA_PARTNERS, OAEBU_METADATA_PARTNERS

MOCK_DATA_PARTNERS = {
    "data_partner": OaebuPartner("data_partner", "dataset1", "table1", "isbn1", "title1", False, "schema_folder1")
}
MOCK_METADATA_PARTNERS = {
    "md_partner": OaebuPartner("md_partner", "dataset1", "table1", "isbn1", "title1", False, "schema_folder1")
}


class TestPartnerFromStr(unittest.TestCase):
    @patch.dict("oaebu_workflows.oaebu_partners.OAEBU_METADATA_PARTNERS", MOCK_METADATA_PARTNERS)
    @patch.dict("oaebu_workflows.oaebu_partners.OAEBU_DATA_PARTNERS", MOCK_DATA_PARTNERS)
    def test_valid_partner_name_string(self):
        # Call the function with a valid partner name string
        result = partner_from_str("data_partner")
        self.assertIsInstance(result, OaebuPartner)
        result = partner_from_str("md_partner", metadata_partner=True)
        self.assertIsInstance(result, OaebuPartner)
        result = partner_from_str(OAEBU_DATA_PARTNERS["data_partner"])
        self.assertEqual(result, OAEBU_DATA_PARTNERS["data_partner"])
        result = partner_from_str(OAEBU_METADATA_PARTNERS["md_partner"], metadata_partner=True)
        self.assertEqual(result, OAEBU_METADATA_PARTNERS["md_partner"])

    @patch.dict("oaebu_workflows.oaebu_partners.OAEBU_METADATA_PARTNERS", MOCK_METADATA_PARTNERS)
    @patch.dict("oaebu_workflows.oaebu_partners.OAEBU_DATA_PARTNERS", MOCK_DATA_PARTNERS)
    def test_invalid_partner_name_string(self):
        # Call the function with an invalid partner name string
        with self.assertRaisesRegex(KeyError, "Partner not found: invalid_partner"):
            partner_from_str("invalid_partner")
        with self.assertRaisesRegex(KeyError, "Partner not found: invalid_partner"):
            partner_from_str("invalid_partner", metadata_partner=True)
        # Call the function with a valid metadata partner but without the flag
        with self.assertRaisesRegex(KeyError, "Partner not found: md_partner"):
            partner_from_str("md_partner", metadata_partner=False)



# Copyright 2020-2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Author: Tuan Chien, Keegan Smith

import os
from typing import Union
from dataclasses import dataclass

from oaebu_workflows.config import schema_folder


@dataclass
class OaebuPartner:
    """Class for storing information about data sources we are using to produce oaebu intermediate tables for.

    :param type_id: The dataset type id. Should be the same as its dictionary key
    :param bq_dataset_id: The BigQuery dataset ID Bigquery Dataset ID.
    :param bq_table_name: The BigQuery table name Bigquery Table name
    :param isbn_field_name: Name of the field containing the ISBN.
    :param title_field_name: Name of the field containing the Title.
    :param sharded: whether the table is sharded or not.
    """

    type_id: str
    bq_dataset_id: str
    bq_table_name: str
    isbn_field_name: str
    title_field_name: str
    sharded: bool
    schema_path: str

    def __str__(self):
        return self.type_id


OAEBU_METADATA_PARTNERS = dict(
    onix=OaebuPartner(
        type_id="onix",
        bq_dataset_id="onix",
        bq_table_name="onix",
        isbn_field_name="ISBN13",
        title_field_name="TitleDetails.TitleElements.TitleText",
        sharded=True,
        schema_path=os.path.join(schema_folder(workflow_module="onix_telescope"), "onix.json"),
    ),
    thoth=OaebuPartner(
        type_id="thoth",
        bq_dataset_id="onix",
        bq_table_name="onix",
        isbn_field_name="ISBN13",
        title_field_name="TitleDetails.TitleElements.TitleText",
        sharded=True,
        schema_path=os.path.join(schema_folder(workflow_module="onix_telescope"), "onix.json"),
    ),
    oapen_metadata=OaebuPartner(
        type_id="oapen_metadata",
        bq_dataset_id="onix",
        bq_table_name="onix",
        isbn_field_name="ISBN13",
        title_field_name="TitleDetails.TitleElements.TitleText",
        sharded=True,
        schema_path=os.path.join(schema_folder(workflow_module="onix_telescope"), "onix.json"),
    ),
)

OAEBU_DATA_PARTNERS = dict(
    google_analytics3=OaebuPartner(
        type_id="google_analytics3",
        bq_dataset_id="google",
        bq_table_name="google_analytics3",
        isbn_field_name="publication_id",
        title_field_name="title",
        sharded=False,
        schema_path=os.path.join(schema_folder(workflow_module="google_analytics3_telescope"), "google_analytics.json"),
    ),
    google_books_sales=OaebuPartner(
        type_id="google_books_sales",
        bq_dataset_id="google",
        bq_table_name="google_books_sales",
        isbn_field_name="Primary_ISBN",
        title_field_name="Title",
        sharded=False,
        schema_path=os.path.join(schema_folder(workflow_module="google_books_telescope"), "google_books_sales.json"),
    ),
    google_books_traffic=OaebuPartner(
        type_id="google_books_traffic",
        bq_dataset_id="google",
        bq_table_name="google_books_traffic",
        isbn_field_name="Primary_ISBN",
        title_field_name="Title",
        sharded=False,
        schema_path=os.path.join(schema_folder(workflow_module="google_books_telescope"), "google_books_traffic.json"),
    ),
    jstor_country=OaebuPartner(
        type_id="jstor_country",
        bq_dataset_id="jstor",
        bq_table_name="jstor_country",
        isbn_field_name="ISBN",
        title_field_name="Book_Title",
        sharded=False,
        schema_path=os.path.join(schema_folder(workflow_module="jstor_telescope"), "jstor_country.json"),
    ),
    jstor_institution=OaebuPartner(
        type_id="jstor_institution",
        bq_dataset_id="jstor",
        bq_table_name="jstor_institution",
        isbn_field_name="ISBN",
        title_field_name="Book_Title",
        sharded=False,
        schema_path=os.path.join(schema_folder(workflow_module="jstor_telescope"), "jstor_institution.json"),
    ),
    jstor_country_collection=OaebuPartner(
        type_id="jstor_country_collection",
        bq_dataset_id="jstor",
        bq_table_name="jstor_country",
        isbn_field_name="ISBN",
        title_field_name="Book_Title",
        sharded=False,
        schema_path=os.path.join(schema_folder(workflow_module="jstor_telescope"), "jstor_country_collection.json"),
    ),
    jstor_institution_collection=OaebuPartner(
        type_id="jstor_institution_collection",
        bq_dataset_id="jstor",
        bq_table_name="jstor_institution",
        isbn_field_name="ISBN",
        title_field_name="Book_Title",
        sharded=False,
        schema_path=os.path.join(schema_folder(workflow_module="jstor_telescope"), "jstor_institution_collection.json"),
    ),
    irus_oapen=OaebuPartner(
        type_id="irus_oapen",
        bq_dataset_id="irus",
        bq_table_name="irus_oapen",
        isbn_field_name="ISBN",
        title_field_name="book_title",
        sharded=False,
        schema_path=os.path.join(schema_folder(workflow_module="irus_oapen_telescope"), "irus_oapen.json"),
    ),
    irus_fulcrum=OaebuPartner(
        type_id="irus_fulcrum",
        bq_dataset_id="irus",
        bq_table_name="irus_fulcrum",
        isbn_field_name="ISBN",
        title_field_name="book_title",
        sharded=False,
        schema_path=os.path.join(schema_folder(workflow_module="irus_fulcrum_telescope"), "irus_fulcrum.json"),
    ),
    ucl_discovery=OaebuPartner(
        type_id="ucl_discovery",
        bq_dataset_id="ucl",
        bq_table_name="ucl_discovery",
        isbn_field_name="ISBN",
        title_field_name="title",
        sharded=False,
        schema_path=os.path.join(schema_folder(workflow_module="ucl_discovery_telescope"), "ucl_discovery.json"),
    ),
    internet_archive=OaebuPartner(
        type_id="internet_archive",
        bq_dataset_id="internet_archive",
        bq_table_name="internet_archive",
        isbn_field_name="ISBN13",
        title_field_name="title",
        sharded=False,
        schema_path=os.path.join(schema_folder(), "internet_archive", "internet_archive.json"),
    ),
    worldreader=OaebuPartner(
        type_id="worldreader",
        bq_dataset_id="worldreader",
        bq_table_name="worldreader",
        isbn_field_name="ISBN13",
        title_field_name="title",
        sharded=False,
        schema_path=os.path.join(schema_folder(), "worldreader", "worldreader.json"),
    ),
)


def partner_from_str(partner: Union[str, OaebuPartner], metadata_partner: bool = False) -> OaebuPartner:
    """Get the partner from a string.

    :param partner: The partner name.
    :param metadata_partner: If True, use the metadata partner dictionary; otherwise, use the data partners dictionary
    :raises Exception: Raised if the partner name is not found
    :return: The OaebuPartner
    """

    if isinstance(partner, OaebuPartner):
        return partner

    partners_dict = OAEBU_METADATA_PARTNERS if metadata_partner else OAEBU_DATA_PARTNERS

    try:
        partner = partners_dict[str(partner)]
    except KeyError as e:
        raise KeyError(f"Partner not found: {partner}").with_traceback(e.__traceback__)

    return partner



# Copyright 2020-2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs, Tuan Chien, Keegan Smith

import os
import unittest
from unittest.mock import MagicMock
from tempfile import NamedTemporaryFile
from xml.parsers.expat import ExpatError

import pendulum
import vcr
from airflow.exceptions import AirflowException
from airflow.utils.state import State
from tenacity import stop_after_attempt

from oaebu_workflows.config import test_fixtures_folder
from oaebu_workflows.oaebu_partners import partner_from_str
from oaebu_workflows.oapen_metadata_telescope.oapen_metadata_telescope import (
    OapenMetadataTelescope,
    download_metadata,
)
from observatory.platform.api import get_dataset_releases
from observatory.platform.observatory_config import Workflow
from observatory.platform.gcs import gcs_blob_name_from_path
from observatory.platform.bigquery import bq_sharded_table_id
from observatory.platform.observatory_environment import (
    ObservatoryEnvironment,
    ObservatoryTestCase,
    find_free_port,
    load_and_parse_json,
)


class TestOapenMetadataTelescope(ObservatoryTestCase):
    """Tests for the Oapen Metadata Telescope DAG"""

    def __init__(self, *args, **kwargs):
        """Constructor which sets up variables used by tests.

        :param args: arguments.
        :param kwargs: keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self.project_id = os.getenv("TEST_GCP_PROJECT_ID")
        self.data_location = os.getenv("TEST_GCP_DATA_LOCATION")
        self.metadata_uri = "https://library.oapen.org/download-export?format=onix"  # metadata URI

        fixtures_folder = test_fixtures_folder(workflow_module="oapen_metadata_telescope")
        self.valid_download_cassette = os.path.join(fixtures_folder, "cassette_valid.yaml")  # VCR Cassette
        self.test_table = os.path.join(fixtures_folder, "test_table.json")  # File for testing final table

    def test_dag_structure(self):
        """Test that the Oapen Metadata DAG has the correct structure"""
        dag = OapenMetadataTelescope(
            dag_id="oapen_metadata",
            cloud_workspace=self.fake_cloud_workspace,
            metadata_uri="",
        ).make_dag()
        self.assert_dag_structure(
            {
                "check_dependencies": ["download"],
                "download": ["upload_downloaded"],
                "upload_downloaded": ["transform"],
                "transform": ["upload_transformed"],
                "upload_transformed": ["bq_load"],
                "bq_load": ["add_new_dataset_releases"],
                "add_new_dataset_releases": ["cleanup"],
                "cleanup": [],
            },
            dag,
        )

    def test_dag_load(self):
        """Test that the OapenMetadata DAG can be loaded from a DAG bag"""
        env = ObservatoryEnvironment(
            workflows=[
                Workflow(
                    dag_id="oapen_metadata",
                    name="OAPEN Metadata Telescope",
                    class_name="oaebu_workflows.oapen_metadata_telescope.oapen_metadata_telescope.OapenMetadataTelescope",
                    cloud_workspace=self.fake_cloud_workspace,
                    kwargs=dict(metadata_uri=""),
                )
            ],
        )
        with env.create():
            self.assert_dag_load_from_config("oapen_metadata")

    def test_telescope(self):
        """Test telescope task execution."""

        env = ObservatoryEnvironment(
            self.project_id, self.data_location, api_host="localhost", api_port=find_free_port()
        )
        dataset_id = env.add_dataset()

        with env.create():
            partner = partner_from_str("oapen_metadata", metadata_partner=True)
            partner.bq_dataset_id = dataset_id
            telescope = OapenMetadataTelescope(
                dag_id="oapen_metadata",
                cloud_workspace=env.cloud_workspace,
                metadata_uri=self.metadata_uri,
                metadata_partner=partner,
                elevate_related_products=True,
                bq_dataset_id=dataset_id,
            )
            dag = telescope.make_dag()

            # first run
            with env.create_dag_run(dag, pendulum.datetime(year=2021, month=2, day=1)):
                # Test that all dependencies are specified: no error should be thrown
                ti = env.run_task(telescope.check_dependencies.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Download task
                with vcr.VCR().use_cassette(self.valid_download_cassette, record_mode="None"):
                    ti = env.run_task(telescope.download.__name__)
                    self.assertEqual(ti.state, State.SUCCESS)

                # Upload download task
                ti = env.run_task(telescope.upload_downloaded.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Transform task
                ti = env.run_task(telescope.transform.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Upload transform task
                ti = env.run_task(telescope.upload_transformed.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Bigquery load task
                ti = env.run_task(telescope.bq_load.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                ### Make Assertions ###

                # Create the release
                release = telescope.make_release(
                    run_id=env.dag_run.run_id, data_interval_end=pendulum.parse(str(env.dag_run.data_interval_end))
                )

                # Test download task
                self.assertTrue(os.path.exists(release.download_path))
                self.assert_file_integrity(release.download_path, "c246a8f7487de756f4dd47cd0ab94363", "md5")

                # Test that download file uploaded to BQ
                self.assert_blob_integrity(
                    env.download_bucket, gcs_blob_name_from_path(release.download_path), release.download_path
                )

                # Test transform task produced the files we care about
                invalid_products_path = os.path.join(release.transform_folder, "invalid_products.xml")
                self.assertTrue(os.path.exists(invalid_products_path))

                # Check file content is as expected
                self.assert_file_integrity(invalid_products_path, "1ce5155e79ff4e405564038d4520ae3c", "md5")

                # Test that transformed files uploaded to BQ
                self.assert_blob_integrity(
                    env.transform_bucket, gcs_blob_name_from_path(release.transform_path), release.transform_path
                )
                self.assert_blob_integrity(
                    env.transform_bucket, gcs_blob_name_from_path(invalid_products_path), invalid_products_path
                )

                # Test that table is loaded to BQ
                table_id = bq_sharded_table_id(
                    telescope.cloud_workspace.project_id,
                    telescope.metadata_partner.bq_dataset_id,
                    telescope.metadata_partner.bq_table_name,
                    release.snapshot_date,
                )
                self.assert_table_integrity(table_id, expected_rows=5)
                self.assert_table_content(table_id, load_and_parse_json(self.test_table), primary_key="ISBN13")

                # Add_dataset_release_task
                dataset_releases = get_dataset_releases(dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id)
                self.assertEqual(len(dataset_releases), 0)
                ti = env.run_task(telescope.add_new_dataset_releases.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                dataset_releases = get_dataset_releases(dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id)
                self.assertEqual(len(dataset_releases), 1)

                # Test that all data deleted
                ti = env.run_task(telescope.cleanup.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assert_cleanup(release.workflow_folder)


class TestDownloadMetadata(unittest.TestCase):
    # Cassettes
    fixtures_folder = test_fixtures_folder(workflow_module="oapen_metadata_telescope")
    valid_download_cassette = os.path.join(fixtures_folder, "cassette_valid.yaml")
    invalid_download_cassette = os.path.join(fixtures_folder, "cassette_invalid.yaml")
    empty_download_cassette = os.path.join(fixtures_folder, "cassette_empty.yaml")
    bad_response_cassette = os.path.join(fixtures_folder, "cassette_bad_response.yaml")
    header_only_download_cassette = os.path.join(fixtures_folder, "cassette_header_only.yaml")

    # XMLs
    valid_download_xml = os.path.join(fixtures_folder, "metadata_download_valid.xml")

    # Download URI
    uri = "https://library.oapen.org/download-export?format=onix"

    # Remove the wait time before retries for testing
    download_metadata.retry.sleep = MagicMock()

    def test_download_metadata(self):
        """Test that metadata successfully downloads after 200 respose"""
        with vcr.VCR().use_cassette(self.valid_download_cassette, record_mode="none", allow_playback_repeats=True):
            with NamedTemporaryFile() as download_file:
                download_metadata(self.uri, download_file.name)
                with open(download_file.name, "r") as f:
                    downloaded_xml = f.readlines()
        with open(self.valid_download_xml, "r") as f:  # Note - do not format this file or this test will fail
            assertion_xml = f.readlines()
        self.assertEqual(len(downloaded_xml), len(assertion_xml))
        self.assertEqual(downloaded_xml, assertion_xml)

    def test_download_metadata_invalid_xml(self):
        """Test behaviour when the downloaded file is an invalid XML"""
        download_metadata.retry.stop = stop_after_attempt(1)
        with vcr.VCR().use_cassette(self.invalid_download_cassette, record_mode="none", allow_playback_repeats=True):
            with NamedTemporaryFile() as download_file:
                self.assertRaises(ExpatError, download_metadata, self.uri, download_file.name)

    def test_download_metadata_empty_xml(self):
        """Test behaviour when the downloaded file is an empty XML"""
        download_metadata.retry.stop = stop_after_attempt(1)
        with vcr.VCR().use_cassette(self.empty_download_cassette, record_mode="none", allow_playback_repeats=True):
            with NamedTemporaryFile() as download_file:
                self.assertRaises(ExpatError, download_metadata, self.uri, download_file.name)

    def test_download_metadata_no_products(self):
        """Test behaviour when the downloaded file is an empty XML"""
        download_metadata.retry.stop = stop_after_attempt(1)
        # For only-header XML
        with vcr.VCR().use_cassette(
            self.header_only_download_cassette, record_mode="none", allow_playback_repeats=True
        ):
            with NamedTemporaryFile() as download_file:
                self.assertRaisesRegex(
                    AirflowException, "No products found", download_metadata, self.uri, download_file.name
                )

    def test_download_metadata_bad_response(self):
        """Test behaviour when the downloaded file has a non-200 response code"""
        download_metadata.retry.stop = stop_after_attempt(1)
        with vcr.VCR().use_cassette(self.bad_response_cassette, record_mode="none", allow_playback_repeats=True):
            with NamedTemporaryFile() as download_file:
                self.assertRaisesRegex(
                    ConnectionError, "Expected status code 200", download_metadata, self.uri, download_file.name
                )



# Copyright 2020-2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs

import csv
import os
import re
from collections import OrderedDict, defaultdict
from typing import List, Tuple, Union

import pendulum
from airflow.exceptions import AirflowException
from airflow.models.taskinstance import TaskInstance
from google.cloud.bigquery import TimePartitioningType, SourceFormat, WriteDisposition

from oaebu_workflows.oaebu_partners import OaebuPartner, partner_from_str
from observatory.api.client.model.dataset_release import DatasetRelease
from observatory.platform.api import make_observatory_api
from observatory.platform.airflow import AirflowConns
from observatory.platform.files import save_jsonl_gz
from observatory.platform.files import convert, add_partition_date
from observatory.platform.gcs import gcs_upload_files, gcs_blob_uri, gcs_blob_name_from_path
from observatory.platform.observatory_config import CloudWorkspace
from observatory.platform.bigquery import bq_load_table, bq_table_id, bq_create_dataset
from observatory.platform.sftp import SftpFolders, make_sftp_connection
from observatory.platform.workflows.workflow import (
    PartitionRelease,
    Workflow,
    cleanup,
    set_task_state,
    check_workflow_inputs,
)


class GoogleBooksRelease(PartitionRelease):
    def __init__(
        self,
        dag_id: str,
        run_id: str,
        partition_date: pendulum.DateTime,
        sftp_files: List[str],
    ):
        """Construct a GoogleBooksRelease.

        :param dag_id: The ID of the DAG
        :param run_id: The Airflow run ID
        :param partition_date: the partition date, corresponds to the last day of the month being processed.
        :param sftp_files: List of full filepaths to download from sftp service (incl. in_progress folder)
        """
        super().__init__(dag_id=dag_id, run_id=run_id, partition_date=partition_date)
        self.download_sales_path = os.path.join(self.download_folder, "google_books_sales.csv")
        self.download_traffic_path = os.path.join(self.download_folder, "google_books_traffic.csv")
        self.transform_sales_path = os.path.join(self.transform_folder, "google_books_sales.jsonl.gz")
        self.transform_traffic_path = os.path.join(self.transform_folder, "google_books_traffic.jsonl.gz")
        self.sftp_files = sftp_files


class GoogleBooksTelescope(Workflow):
    """The Google Books telescope."""

    def __init__(
        self,
        dag_id: str,
        cloud_workspace: CloudWorkspace,
        sftp_root: str = "/",
        sales_partner: Union[str, OaebuPartner] = "google_books_sales",
        traffic_partner: Union[str, OaebuPartner] = "google_books_traffic",
        bq_dataset_description: str = "Data from Google sources",
        bq_sales_table_description: str = None,
        bq_traffic_table_description: str = None,
        api_dataset_id: str = "google_books",
        sftp_service_conn_id: str = "sftp_service",
        observatory_api_conn_id: str = AirflowConns.OBSERVATORY_API,
        catchup: bool = False,
        schedule: str = "@weekly",
        start_date: pendulum.DateTime = pendulum.datetime(2018, 1, 1),
    ):
        """Construct a GoogleBooksTelescope instance.
        :param dag_id: The ID of the DAG
        :param cloud_workspace: The CloudWorkspace object for this DAG
        :param sftp_root: The root of the SFTP filesystem to work with
        :param sales_partner: The name of the sales partner
        :param traffic_partner: The name of the traffic partner
        :param bq_dataset_description: Description for the BigQuery dataset
        :param bq_sales_table_description: Description for the BigQuery Google Books Sales table
        :param bq_traffic_table_description: Description for the BigQuery Google Books Traffic table
        :param api_dataset_id: The ID to store the dataset release in the API
        :param sftp_service_conn_id: Airflow connection ID for the SFTP service
        :param observatory_api_conn_id: Airflow connection ID for the overvatory API
        :param catchup: Whether to catchup the DAG or not
        :param schedule: The schedule interval of the DAG
        :param start_date: The start date of the DAG
        """
        super().__init__(
            dag_id,
            start_date,
            schedule,
            catchup=catchup,
            airflow_conns=[sftp_service_conn_id, observatory_api_conn_id],
            tags=["oaebu"],
        )
        self.dag_id = dag_id
        self.cloud_workspace = cloud_workspace
        self.sftp_root = sftp_root
        self.sales_partner = partner_from_str(sales_partner)
        self.traffic_partner = partner_from_str(traffic_partner)
        self.bq_dataset_description = bq_dataset_description
        self.bq_sales_table_description = bq_sales_table_description
        self.bq_traffic_table_description = bq_traffic_table_description
        self.api_dataset_id = api_dataset_id
        self.sftp_service_conn_id = sftp_service_conn_id
        self.observatory_api_conn_id = observatory_api_conn_id

        # Extra SFTP parameters
        self.sftp_folders = SftpFolders(dag_id, sftp_conn_id=sftp_service_conn_id, sftp_root=sftp_root)
        self.sftp_regex = r"^Google(SalesTransaction|BooksTraffic)Report_\d{4}_\d{2}.csv$"

        check_workflow_inputs(self)

        self.add_setup_task(self.check_dependencies)
        self.add_setup_task(self.list_release_info)
        self.add_task(self.move_files_to_in_progress)
        self.add_task(self.download)
        self.add_task(self.upload_downloaded)
        self.add_task(self.transform)
        self.add_task(self.upload_transformed)
        self.add_task(self.bq_load)
        self.add_task(self.move_files_to_finished)
        self.add_task(self.add_new_dataset_releases)
        self.add_task(self.cleanup)

    def make_release(self, **kwargs) -> List[GoogleBooksRelease]:
        """Make release instances. The release is passed as an argument to the function (TelescopeFunction) that is
        called in 'task_callable'.

        :param kwargs: the context passed from the PythonOperator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for the keyword arguments that can be passed
        :return: A list of google books release instances
        """
        ti: TaskInstance = kwargs["ti"]
        reports_info = ti.xcom_pull(
            key=GoogleBooksTelescope.RELEASE_INFO, task_ids=self.list_release_info.__name__, include_prior_dates=False
        )
        releases = []
        run_id = kwargs["run_id"]
        for partition_date, sftp_files in reports_info.items():
            releases.append(
                GoogleBooksRelease(
                    self.dag_id, run_id=run_id, partition_date=pendulum.parse(partition_date), sftp_files=sftp_files
                )
            )
        return releases

    def list_release_info(self, **kwargs) -> bool:
        """Lists all Google Books releases available on the SFTP server and publishes sftp file paths and
        release_date's as an XCom.

        :return: the identifier of the task to execute next.
        """

        reports = defaultdict(list)
        # List all reports in the 'upload' folder of the organisation
        with make_sftp_connection(self.sftp_service_conn_id) as sftp:
            files = sftp.listdir(self.sftp_folders.upload)
            for file_name in files:
                match = re.match(self.sftp_regex, file_name)
                if match:
                    # Get the release date from file name
                    date_str = file_name[-11:].strip(".csv")
                    release_date = pendulum.from_format(date_str, "YYYY_MM").end_of("month")
                    release_date = release_date.format("YYYYMMDD")

                    # Get the report type from file name
                    report_type = match.group(1)

                    # Create the full path of the file for the 'in progress' folder
                    sftp_file = os.path.join(self.sftp_folders.in_progress, file_name)

                    # Append report
                    reports[report_type + release_date].append(sftp_file)

        # Check that for each report type + date combination there is a report available
        release_info = defaultdict(list)
        for report, sftp_files in reports.items():
            release_date = report[-8:]
            release_info[release_date] += sftp_files

        continue_dag = bool(release_info)
        if continue_dag:
            # Push messages
            ti: TaskInstance = kwargs["ti"]
            ti.xcom_push(GoogleBooksTelescope.RELEASE_INFO, release_info)

        return continue_dag

    def move_files_to_in_progress(self, releases: List[GoogleBooksRelease], **kwargs) -> None:
        """Move Google Books files to SFTP in-progress folder."""

        for release in releases:
            self.sftp_folders.move_files_to_in_progress(release.sftp_files)

    def download(self, releases: List[GoogleBooksRelease], **kwargs):
        """Task to download the Google Books releases for a given month."""
        for release in releases:
            with make_sftp_connection(self.sftp_service_conn_id) as sftp:
                for file in release.sftp_files:
                    if "Traffic" in file:
                        sftp.get(file, localpath=release.download_traffic_path)
                    elif "Transaction" in file:
                        sftp.get(file, localpath=release.download_sales_path)
            assert os.path.exists(release.download_traffic_path) and os.path.exists(release.download_sales_path)

    def upload_downloaded(self, releases: List[GoogleBooksRelease], **kwargs) -> None:
        """Uploads the downloaded files to GCS for each release"""
        for release in releases:
            success = gcs_upload_files(
                bucket_name=self.cloud_workspace.download_bucket,
                file_paths=[release.download_sales_path, release.download_traffic_path],
            )
            if not success:
                raise AirflowException(f"Files could not be uploaded to cloud storage bucket: {self.transform_bucket}")

    def transform(self, releases: List[GoogleBooksRelease], **kwargs) -> None:
        """Task to transform the Google Books releases for a given month."""
        for release in releases:
            gb_transform(
                download_files=(release.download_sales_path, release.download_traffic_path),
                sales_path=release.transform_sales_path,
                traffic_path=release.transform_traffic_path,
                release_date=release.partition_date,
            )

    def upload_transformed(self, releases: List[GoogleBooksRelease], **kwargs) -> None:
        """Uploads the transformed files to GCS for each release"""
        for release in releases:
            success = gcs_upload_files(
                bucket_name=self.cloud_workspace.transform_bucket,
                file_paths=[release.transform_sales_path, release.transform_traffic_path],
            )
            if not success:
                raise AirflowException(f"Files could not be uploaded to cloud storage bucket: {self.transform_bucket}")

    def move_files_to_finished(self, releases: List[GoogleBooksRelease], **kwargs) -> None:
        """Move Google Books files to SFTP finished folder."""

        for release in releases:
            self.sftp_folders.move_files_to_finished(release.sftp_files)

    def bq_load(self, releases: List[GoogleBooksRelease], **kwargs) -> None:
        """Loads the sales and traffic data into BigQuery"""
        for release in releases:
            for partner, table_description, file_path in [
                [self.sales_partner, self.bq_sales_table_description, release.transform_sales_path],
                [self.traffic_partner, self.bq_traffic_table_description, release.transform_traffic_path],
            ]:
                bq_create_dataset(
                    project_id=self.cloud_workspace.project_id,
                    dataset_id=partner.bq_dataset_id,
                    location=self.cloud_workspace.data_location,
                    description=self.bq_dataset_description,
                )
                uri = gcs_blob_uri(self.cloud_workspace.transform_bucket, gcs_blob_name_from_path(file_path))
                table_id = bq_table_id(self.cloud_workspace.project_id, partner.bq_dataset_id, partner.bq_table_name)
                success = bq_load_table(
                    uri=uri,
                    table_id=table_id,
                    schema_file_path=partner.schema_path,
                    source_format=SourceFormat.NEWLINE_DELIMITED_JSON,
                    partition_type=TimePartitioningType.MONTH,
                    partition=True,
                    partition_field="release_date",
                    write_disposition=WriteDisposition.WRITE_APPEND,
                    table_description=table_description,
                    ignore_unknown_values=True,
                )
                set_task_state(success, kwargs["ti"].task_id, release=release)

    def add_new_dataset_releases(self, releases: List[GoogleBooksRelease], **kwargs) -> None:
        """Adds release information to API."""

        api = make_observatory_api(observatory_api_conn_id=self.observatory_api_conn_id)
        for release in releases:
            dataset_release = DatasetRelease(
                dag_id=self.dag_id,
                dataset_id=self.api_dataset_id,
                dag_run_id=release.run_id,
                data_interval_start=kwargs["data_interval_start"],
                data_interval_end=kwargs["data_interval_end"],
                partition_date=release.partition_date,
            )
            api.post_dataset_release(dataset_release)

    def cleanup(self, releases: List[GoogleBooksRelease], **kwargs) -> None:
        """Delete all files, folders and XComs associated with this release."""
        for release in releases:
            cleanup(
                dag_id=self.dag_id, execution_date=kwargs["execution_date"], workflow_folder=release.workflow_folder
            )


def gb_transform(
    download_files: Tuple[str, str], sales_path: str, traffic_path: str, release_date: pendulum.DateTime
) -> None:
    """Transforms sales and traffic reports. For both reports it transforms the csv into a jsonl file and
    replaces spaces in the keys with underscores.

    :param download_files: The Google Books Sales and Traffic files
    :param sales_path: The file path to save the transformed sales data to
    :param traffic_path: The file path to save the transformed traffic data to
    :param release_date: The release date to use as a partitioning date
    """
    # Sort files to get same hash for unit tests

    results = defaultdict(list)
    results["sales"] = []
    results["traffic"] = []
    for file in download_files:
        report_type = "sales" if "sales" in os.path.basename(file).lower() else "traffic"
        with open(file, encoding="utf-16") as csv_file:
            csv_reader = csv.DictReader(csv_file, delimiter="\t")
            for row in csv_reader:
                transformed_row = OrderedDict((convert(k.replace("%", "Perc")), v) for k, v in row.items())
                # Sales transaction report
                if report_type == "sales":
                    transaction_date = pendulum.from_format(transformed_row["Transaction_Date"], "MM/DD/YY")

                    # Sanity check that transaction date is in month of release date
                    if release_date.start_of("month") <= transaction_date <= release_date.end_of("month"):
                        pass
                    else:
                        raise AirflowException(
                            "Transaction date does not fall within release month. "
                            f"Transaction date: {transaction_date.strftime('%Y-%m-%d')}, "
                            f"release month: {release_date.strftime('%Y-%m')}"
                        )

                    # Transform to valid date format
                    transformed_row["Transaction_Date"] = transaction_date.strftime("%Y-%m-%d")

                    # Remove percentage sign
                    transformed_row["Publisher_Revenue_Perc"] = transformed_row["Publisher_Revenue_Perc"].strip("%")
                    # This field is not present for some publishers (UCL Press), for ANU Press the field value is
                    # “E-Book”
                    try:
                        transformed_row["Line_of_Business"]
                    except KeyError:
                        transformed_row["Line_of_Business"] = None
                # Traffic report
                else:
                    # Remove percentage sign
                    transformed_row["Buy_Link_CTR"] = transformed_row["Buy_Link_CTR"].strip("%")

                # Append results
                results[report_type].append(transformed_row)

    for report_type, report_results in results.items():
        report_results = add_partition_date(
            report_results,
            partition_date=release_date,
            partition_type=TimePartitioningType.MONTH,
            partition_field="release_date",
        )
        save_path = sales_path if report_type == "sales" else traffic_path
        print(f"SAVING REPORT '{report_type}' to {save_path}")
        save_jsonl_gz(save_path, report_results)



# Copyright 2022-2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Keegan Smith

import logging
import os
from typing import List, Tuple, Union

import pendulum
from airflow.hooks.base import BaseHook
from google.cloud.bigquery import SourceFormat, WriteDisposition
from google.cloud.bigquery.table import TimePartitioningType

from oaebu_workflows.oaebu_partners import OaebuPartner, partner_from_str
from observatory.platform.files import add_partition_date
from observatory.platform.api import make_observatory_api, DatasetRelease
from observatory.platform.airflow import AirflowConns
from observatory.platform.observatory_config import CloudWorkspace
from observatory.platform.files import save_jsonl_gz, load_jsonl
from observatory.platform.gcs import gcs_blob_name_from_path, gcs_upload_files, gcs_blob_uri
from observatory.platform.bigquery import bq_load_table, bq_create_dataset, bq_table_id
from observatory.platform.workflows.workflow import (
    Workflow,
    PartitionRelease,
    cleanup,
    set_task_state,
    check_workflow_inputs,
)
from observatory.platform.utils.url_utils import retry_get_url

IRUS_FULCRUM_ENDPOINT_TEMPLATE = (
    "https://irus.jisc.ac.uk/api/v3/irus/reports/irus_ir/?platform=235"
    "&requestor_id={requestor_id}&begin_date={start_date}&end_date={end_date}"
)


class IrusFulcrumRelease(PartitionRelease):
    def __init__(
        self,
        dag_id: str,
        run_id: str,
        data_interval_start: pendulum.DateTime,
        data_interval_end: pendulum.DateTime,
        partition_date: pendulum.DateTime,
    ):
        """Create a IrusFulcrumRelease instance.

        :param dag_id: The ID of the DAG
        :param run_id: The airflow run ID
        :param data_interval_start: The beginning of the data interval
        :param data_interval_end: The end of the data interval
        :param partition_date: The release/partition date
        """
        super().__init__(dag_id=dag_id, run_id=run_id, partition_date=partition_date)
        self.data_interval_start = data_interval_start
        self.data_interval_end = data_interval_end
        self.download_totals_path = os.path.join(self.download_folder, "fulcrum_totals.jsonl.gz")
        self.download_country_path = os.path.join(self.download_folder, "fulcrum_country.json.gz")
        self.transform_path = os.path.join(self.transform_folder, "fulcrum.json.gz")


class IrusFulcrumTelescope(Workflow):
    def __init__(
        self,
        dag_id: str,
        cloud_workspace: CloudWorkspace,
        publishers: List[str],
        data_partner: Union[str, OaebuPartner] = "irus_fulcrum",
        bq_dataset_description: str = "IRUS dataset",
        bq_table_description: str = None,
        api_dataset_id: str = "fulcrum",
        observatory_api_conn_id: str = AirflowConns.OBSERVATORY_API,
        irus_oapen_api_conn_id: str = "irus_api",
        catchup: bool = True,
        schedule: str = "0 0 4 * *",  # Run on the 4th of every month
        start_date: pendulum.DateTime = pendulum.datetime(2022, 4, 1),  # Earliest available data
    ):
        """The Fulcrum Telescope
        :param dag_id: The ID of the DAG
        :param cloud_workspace: The CloudWorkspace object for this DAG
        :param publishers: The publishers pertaining to this DAG instance (as listed in Fulcrum)
        :param data_partner: The name of the data partner
        :param bq_dataset_description: Description for the BigQuery dataset
        :param bq_table_description: Description for the biguery table
        :param api_dataset_id: The ID to store the dataset release in the API
        :param observatory_api_conn_id: Airflow connection ID for the overvatory API
        :param irus_oapen_api_conn_id: Airflow connection ID OAPEN IRUS UK (counter 5)
        :param catchup: Whether to catchup the DAG or not
        :param schedule: The schedule interval of the DAG
        :param start_date: The start date of the DAG
        """
        if bq_table_description is None:
            bq_table_description = "Fulcrum metrics as recorded by the IRUS platform"

        super().__init__(
            dag_id,
            start_date,
            schedule,
            airflow_conns=[observatory_api_conn_id, irus_oapen_api_conn_id],
            catchup=catchup,
            tags=["oaebu"],
        )

        self.dag_id = dag_id
        self.cloud_workspace = cloud_workspace
        self.publishers = publishers
        self.data_partner = partner_from_str(data_partner)
        self.bq_dataset_description = bq_dataset_description
        self.bq_table_description = bq_table_description
        self.api_dataset_id = api_dataset_id
        self.observatory_api_conn_id = observatory_api_conn_id
        self.irus_oapen_api_conn_id = irus_oapen_api_conn_id

        check_workflow_inputs(self)

        self.add_setup_task(self.check_dependencies)
        self.add_task(self.download)
        self.add_task(self.upload_downloaded)
        self.add_task(self.transform)
        self.add_task(self.upload_transformed)
        self.add_task(self.bq_load)
        self.add_task(self.add_new_dataset_releases)
        self.add_task(self.cleanup)

    def make_release(self, **kwargs) -> IrusFulcrumRelease:
        """Create a IrusFulcrumRelease instance
        Dates are best explained with an example
        Say the dag is scheduled to run on 2022-04-07
        Interval_start will be 2022-03-01
        Interval_end will be 2022-04-01
        partition_date will be 2022-03-31
        """
        data_interval_start = kwargs["data_interval_start"].start_of("month")
        data_interval_end = kwargs["data_interval_end"].start_of("month")
        partition_date = data_interval_start.end_of("month")
        return IrusFulcrumRelease(
            self.dag_id,
            kwargs["run_id"],
            data_interval_start=data_interval_start,
            data_interval_end=data_interval_end,
            partition_date=partition_date,
        )

    def download(self, release: IrusFulcrumRelease, **kwargs):
        """Task to download the Fulcrum data for a release

        :param releases: the IrusFulcrumRelease instance.
        """
        requestor_id = BaseHook.get_connection(self.irus_oapen_api_conn_id).login
        totals_data, country_data = download_fulcrum_month_data(release.partition_date, requestor_id)
        assert totals_data and country_data, f"Data not available for supplied release month: {release.partition_date}"
        save_jsonl_gz(release.download_totals_path, totals_data)
        save_jsonl_gz(release.download_country_path, country_data)

    def upload_downloaded(self, release: IrusFulcrumRelease, **kwargs):
        """Upload the downloaded fulcrum data to the google cloud download bucket"""
        success = gcs_upload_files(
            bucket_name=self.cloud_workspace.download_bucket,
            file_paths=[release.download_totals_path, release.download_country_path],
        )
        set_task_state(success, kwargs["ti"].task_id, release=release)

    def transform(self, release: IrusFulcrumRelease, **kwargs):
        """Task to transform the fulcrum data"""
        logging.info(f"Transforming the Fulcrum dataset with the following publisher filter: {self.publishers}")
        totals_data = load_jsonl(release.download_totals_path)
        country_data = load_jsonl(release.download_country_path)
        transformed_data = transform_fulcrum_data(
            totals_data=totals_data,
            country_data=country_data,
            publishers=self.publishers,
        )
        transformed_data = add_partition_date(
            transformed_data,
            partition_date=release.partition_date.end_of("month"),
            partition_type=TimePartitioningType.MONTH,
            partition_field="release_date",
        )
        save_jsonl_gz(release.transform_path, transformed_data)

    def upload_transformed(self, release: IrusFulcrumRelease, **kwargs):
        """Upload the transformed fulcrum data to the google cloud download bucket"""
        success = gcs_upload_files(
            bucket_name=self.cloud_workspace.transform_bucket, file_paths=[release.transform_path]
        )
        set_task_state(success, kwargs["ti"].task_id, release=release)

    def bq_load(self, release: IrusFulcrumRelease, **kwargs) -> None:
        """Load the transfromed data into bigquery"""
        bq_create_dataset(
            project_id=self.cloud_workspace.project_id,
            dataset_id=self.data_partner.bq_dataset_id,
            location=self.cloud_workspace.data_location,
            description=self.bq_dataset_description,
        )

        # Load each transformed release
        uri = gcs_blob_uri(self.cloud_workspace.transform_bucket, gcs_blob_name_from_path(release.transform_path))
        table_id = bq_table_id(
            self.cloud_workspace.project_id, self.data_partner.bq_dataset_id, self.data_partner.bq_table_name
        )
        success = bq_load_table(
            uri=uri,
            table_id=table_id,
            schema_file_path=self.data_partner.schema_path,
            source_format=SourceFormat.NEWLINE_DELIMITED_JSON,
            table_description=self.bq_table_description,
            partition=True,
            partition_type=TimePartitioningType.MONTH,
            write_disposition=WriteDisposition.WRITE_APPEND,
            partition_field="release_date",
            ignore_unknown_values=True,
        )
        set_task_state(success, kwargs["ti"].task_id, release=release)

    def add_new_dataset_releases(self, release: IrusFulcrumRelease, **kwargs) -> None:
        """Adds release information to API."""
        api = make_observatory_api(observatory_api_conn_id=self.observatory_api_conn_id)
        dataset_release = DatasetRelease(
            dag_id=self.dag_id,
            dataset_id=self.api_dataset_id,
            dag_run_id=release.run_id,
            data_interval_start=release.data_interval_start,
            data_interval_end=release.data_interval_end,
            partition_date=release.partition_date,
        )
        api.post_dataset_release(dataset_release)

    def cleanup(self, release: IrusFulcrumRelease, **kwargs) -> None:
        """Delete all files and folders associated with this release."""
        cleanup(self.dag_id, execution_date=kwargs["execution_date"], workflow_folder=release.workflow_folder)


def download_fulcrum_month_data(
    download_month: pendulum.DateTime,
    requestor_id: str,
    num_retries: str = 3,
) -> Tuple[List[dict], List[dict]]:
    """Download Fulcrum data for the release month

    :param download_month: The month to download usage data from
    :param requestor_id: The requestor ID - used to access irus platform
    :param num_retries: Number of attempts to make for the URL
    """
    download_month = download_month.format("YYYY-MM")
    base_url = IRUS_FULCRUM_ENDPOINT_TEMPLATE.format(
        requestor_id=requestor_id,
        start_date=download_month,
        end_date=download_month,
    )
    country_url = base_url + "&attributes_to_show=Country"
    logging.info(f"Downloading Fulcrum metrics for month: {download_month}")
    totals_data = retry_get_url(base_url, num_retries=num_retries).json()
    country_data = retry_get_url(country_url, num_retries=num_retries).json()
    totals_data = totals_data.get("Report_Items")
    country_data = country_data.get("Report_Items")

    return totals_data, country_data


def transform_fulcrum_data(
    totals_data: List[dict],
    country_data: List[dict],
    publishers: List[str] = None,
) -> List[dict]:
    """
    Transforms Fulcrum downloaded "totals" and "country" data.

    :param totals_data: Fulcrum usage data aggregated over all countries
    :param country_data: Fulcrum usage data split by country
    :param publishers: Fulcrum publishers to retain. If None, use all publishers
    """
    # Extract only the publishers related to this organisation name
    if publishers:
        totals_data = [i for i in totals_data if i["Publisher"] in publishers]
        country_data = [i for i in country_data if i["Publisher"] in publishers]

    # Total and Country-granulated results should all have the same item entries and be ordered the same, but we should check anyway
    c_ids = [i["IRUS_Item_ID"] for i in country_data]
    t_ids = [i["IRUS_Item_ID"] for i in totals_data]
    assert len(c_ids) == len(t_ids), "Country entry data is not the same length as total entry data"

    # Mapping the IDs to list elements
    c_id_mapping = {entry["IRUS_Item_ID"]: i for (i, entry) in enumerate(country_data)}
    t_id_mapping = {entry["IRUS_Item_ID"]: i for (i, entry) in enumerate(totals_data)}

    transformed_data = []
    for t_id, c_id in zip(t_ids, c_ids):
        transformed_row = {}
        t_entry = totals_data[t_id_mapping[t_id]]
        c_entry = country_data[c_id_mapping[c_id]]

        # Metrics with country granulation
        country_metrics = []
        for c_metric in c_entry["Performance_Instances"]:  # For each country
            country_metrics.append(
                {
                    "name": c_metric["Country"]["Country"],
                    "code": c_metric["Country"]["Country_Code"],
                    "Total_Item_Investigations": c_metric["Metric_Type_Counts"].get("Total_Item_Investigations"),
                    "Total_Item_Requests": c_metric["Metric_Type_Counts"].get("Total_Item_Requests"),
                    "Unique_Item_Investigations": c_metric["Metric_Type_Counts"].get("Unique_Item_Investigations"),
                    "Unique_Item_Requests": c_metric["Metric_Type_Counts"].get("Unique_Item_Requests"),
                }
            )

        # Total Metrics
        t_metric = t_entry["Performance_Instances"][0]
        total_item_investigations = t_metric["Metric_Type_Counts"].get("Total_Item_Investigations")
        total_item_requests = t_metric["Metric_Type_Counts"].get("Total_Item_Requests")
        unique_item_investigations = t_metric["Metric_Type_Counts"].get("Unique_Item_Investigations")
        unique_item_requests = t_metric["Metric_Type_Counts"].get("Unique_Item_Requests")

        # Row structure
        transformed_row = {
            "proprietary_id": t_id,  # t_id == c_id
            "ISBN": t_entry.get("ISBN"),
            "book_title": t_entry.get("Item"),
            "publisher": t_entry.get("Publisher"),
            "authors": t_entry.get("Authors"),
            "event_month": pendulum.parse(t_entry["Performance_Instances"][0]["Event_Month"]).format("YYYY-MM"),
            "total_item_investigations": total_item_investigations,
            "total_item_requests": total_item_requests,
            "unique_item_investigations": unique_item_investigations,
            "unique_item_requests": unique_item_requests,
            "country": country_metrics,
        }
        transformed_data.append(transformed_row)

    return transformed_data



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Tuan Chien


import unittest

from oaebu_workflows.onix_workflow.onix_work_aggregation import (
    BookWork,
    BookWorkAggregator,
    BookWorkFamily,
    BookWorkFamilyAggregator,
    UnionFind,
    get_pref_product_id,
)


class TestUnionFind(unittest.TestCase):
    """
    Test the UnionFind class.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def test_root_trivial(self):
        n = 5
        uf = UnionFind(n)

        for i in range(n):
            self.assertEqual(uf.root(i), i)

    def test_unite_two(self):
        n = 5
        uf = UnionFind(n)
        uf.unite(0, 1)

        self.assertEqual(uf.root(0), uf.root(1))

    def test_unite_two_parts(self):
        n = 5
        uf = UnionFind(n)
        uf.unite(0, 1)
        uf.unite(2, 3)

        self.assertEqual(uf.root(0), uf.root(1))
        self.assertEqual(uf.root(2), uf.root(3))
        self.assertNotEqual(uf.root(0), uf.root(2))

    def test_unite_two_parts_merge(self):
        n = 5
        uf = UnionFind(n)
        uf.unite(0, 1)
        uf.unite(2, 3)
        uf.unite(0, 2)

        self.assertEqual(uf.root(0), uf.root(1))
        self.assertEqual(uf.root(2), uf.root(3))

        for i in range(4):
            self.assertEqual(uf.root(0), uf.root(i))

    def test_find_trivial(self):
        n = 5
        uf = UnionFind(n)

        for i in range(n - 1):
            self.assertFalse(uf.find(i, i + 1))

    def test_find_after_merge(self):
        n = 5
        uf = UnionFind(n)
        uf.unite(0, 1)

        self.assertTrue(uf.find(0, 1))
        for i in range(1, n - 1):
            self.assertFalse(uf.find(i, i + 1))

    def test_find_after_two_parts_merge(self):
        n = 5
        uf = UnionFind(n)
        uf.unite(0, 1)
        uf.unite(2, 3)
        uf.unite(0, 2)

        self.assertFalse(uf.find(3, 4))
        self.assertTrue(uf.find(0, 3))

    def test_get_partition_trivial(self):
        n = 5
        uf = UnionFind(n)
        partition = uf.get_partition()

        self.assertEqual(len(partition), n)

        for i in range(n):
            self.assertEqual(partition[i][0], i)

    def test_get_partition_two_parts(self):
        n = 5
        uf = UnionFind(n)
        uf.unite(0, 1)
        uf.unite(2, 3)

        partition = uf.get_partition()
        self.assertEqual(len(partition), 3)
        self.assertEqual(partition[0][0], 0)
        self.assertEqual(partition[0][1], 1)
        self.assertEqual(partition[1][0], 2)
        self.assertEqual(partition[1][1], 3)
        self.assertEqual(partition[2][0], 4)

    def test_get_partition_two_parts_merge(self):
        n = 5
        uf = UnionFind(n)
        uf.unite(0, 1)
        uf.unite(2, 3)
        uf.unite(0, 2)

        partition = uf.get_partition()
        self.assertEqual(len(partition), 2)
        self.assertEqual(partition[1][0], 4)

        for i in range(4):
            self.assertEqual(partition[0][i], i)


class TestBookWork(unittest.TestCase):
    """
    Test the BookWork class.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def test_ctor(self):
        work = BookWork(work_id="testid", work_id_type="wid_type", products=[{"ISBN13": "123"}])

        self.assertEqual(work.work_id, "testid")
        self.assertEqual(work.work_id_type, "wid_type")
        self.assertEqual(len(work.products), 1)
        self.assertEqual(len(work.isbns), 1)
        self.assertTrue("123" in work.isbns)

    def test_add_product(self):
        work = BookWork(work_id="testid", work_id_type="wid_type", products=[{"ISBN13": "123"}])

        work.add_product({"ISBN13": "456"})
        self.assertEqual(len(work.products), 2)
        self.assertEqual(len(work.isbns), 2)

        self.assertEqual(work.products[1]["ISBN13"], "456")
        self.assertTrue("456" in work.isbns)

        work.add_product({"ISBN13": "456"})
        self.assertEqual(len(work.products), 2)
        self.assertEqual(len(work.isbns), 2)


class TestBookWorkFamily(unittest.TestCase):
    """
    Test the BookWorkFamily class.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def test_ctor(self):
        work = BookWork(work_id="testid", work_id_type="wid_type", products=[{"ISBN13": "123"}])

        wfam = BookWorkFamily(works=[work])
        self.assertEqual(wfam.work_family_id, None)
        self.assertEqual(wfam.work_family_id_type, None)

        self.assertEqual(len(wfam.works), 1)
        self.assertEqual(wfam.works[0].work_id, "testid")


class TestGetPrefProductId(unittest.TestCase):
    """
    Test the get_pref_product_id function.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def test_unknown(self):
        relprod = {"unknown": "123"}
        self.assertRaises(Exception, get_pref_product_id, relprod)

    def test_get_pidproprietary(self):
        relprod = {"PID_Proprietary": "pidpro"}
        ptype, pid = get_pref_product_id(relprod)
        self.assertEqual(ptype, "PID_Proprietary")
        self.assertEqual(pid, "pidpro")

    def test_get_doi(self):
        relprod = {"PID_Proprietary": "pidpro", "DOI": "doi"}
        ptype, pid = get_pref_product_id(relprod)
        self.assertEqual(ptype, "DOI")
        self.assertEqual(pid, "doi")

    def test_get_gtin13(self):
        relprod = {"PID_Proprietary": "pidpro", "DOI": "doi", "GTIN_13": "gtin"}
        ptype, pid = get_pref_product_id(relprod)
        self.assertEqual(ptype, "GTIN_13")
        self.assertEqual(pid, "gtin")

    def test_get_isbn(self):
        relprod = {"PID_Proprietary": "pidpro", "DOI": "doi", "GTIN_13": "gtin", "ISBN13": "isbn"}
        ptype, pid = get_pref_product_id(relprod)
        self.assertEqual(ptype, "ISBN13")
        self.assertEqual(pid, "isbn")


class TestGetPrefWorkId(unittest.TestCase):
    """
    Test the get_pref_work_id function.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.agg = BookWorkAggregator([])

    def test_unknown(self):
        ids = [{"WorkIDType": "unknown", "IDValue": "value"}]
        wtype, wid = self.agg.get_pref_work_id(ids)

        self.assertEqual(wtype, "unknown")
        self.assertEqual(wid, "value")

    def test_pid_proprietary(self):
        ids = [{"WorkIDType": "unknown", "IDValue": "value"}, {"WorkIDType": "Proprietary", "IDValue": "pidpro"}]

        wtype, wid = self.agg.get_pref_work_id(ids)
        self.assertEqual(wtype, "Proprietary")
        self.assertEqual(wid, "pidpro")

    def test_doi(self):
        ids = [
            {"WorkIDType": "unknown", "IDValue": "value"},
            {"WorkIDType": "PID_Proprietary", "IDValue": "pidpro"},
            {"WorkIDType": "DOI", "IDValue": "doi"},
        ]

        wtype, wid = self.agg.get_pref_work_id(ids)
        self.assertEqual(wtype, "DOI")
        self.assertEqual(wid, "doi")

    def test_gtin13(self):
        ids = [
            {"WorkIDType": "unknown", "IDValue": "value"},
            {"WorkIDType": "PID_Proprietary", "IDValue": "pidpro"},
            {"WorkIDType": "DOI", "IDValue": "doi"},
            {"WorkIDType": "GTIN_13", "IDValue": "gtin"},
        ]

        wtype, wid = self.agg.get_pref_work_id(ids)
        self.assertEqual(wtype, "GTIN_13")
        self.assertEqual(wid, "gtin")

    def test_isbn13(self):
        ids = [
            {"WorkIDType": "unknown", "IDValue": "value"},
            {"WorkIDType": "PID_Proprietary", "IDValue": "pidpro"},
            {"WorkIDType": "DOI", "IDValue": "doi"},
            {"WorkIDType": "GTIN_13", "IDValue": "gtin"},
            {"WorkIDType": "ISBN-13", "IDValue": "isbn"},
        ]

        wtype, wid = self.agg.get_pref_work_id(ids)
        self.assertEqual(wtype, "ISBN13")
        self.assertEqual(wid, "isbn")


class TestBookWorkAggregator(unittest.TestCase):
    """
    Test the BookWorkAggregator class.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def test_ctor(self):
        agg = BookWorkAggregator([])
        self.assertEqual(len(agg.records), 0)

    def test_is_relevant_worK_relation(self):
        agg = BookWorkAggregator([])
        relwork1 = {"WorkRelationCode": "Manifestation of"}

        self.assertTrue(agg.is_relevant_work_relation(relwork1))

    def test_aggregate_empty(self):
        agg = BookWorkAggregator([])
        works = agg.aggregate()
        self.assertEqual(len(works), 0)

    def test_agg_relworks(self):
        agg = BookWorkAggregator(
            [
                {
                    "ISBN13": "123",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [
                                {"WorkIDType": "ISBN-13", "IDValue": "246"},
                                {"WorkIDType": "GTIN_13", "IDValue": "123"},
                            ],
                        },
                    ],
                },
                {
                    "ISBN13": "246",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "ISBN-13", "IDValue": "246"}],
                        },
                    ],
                },
                {
                    "ISBN13": "111",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "ISBN-13", "IDValue": "111"}],
                        },
                        {
                            "WorkRelationCode": "something unknown",
                            "WorkIdentifiers": [{"WorkIDType": "ISBN-13", "IDValue": "246"}],
                        },
                    ],
                },
                {
                    "ISBN13": "222",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "ISBN-13", "IDValue": "222"}],
                        },
                    ],
                },
            ]
        )

        agg.agg_relworks()
        works = agg.get_works_from_partition(agg.uf.get_partition())
        self.assertEqual(len(works), 3)

        work1 = works[0]
        self.assertEqual(len(work1.products), 2)
        self.assertTrue("123" in work1.isbns)
        self.assertTrue("246" in work1.isbns)

        work2 = works[1]
        self.assertEqual(len(work2.products), 1)
        self.assertTrue("111" in work2.isbns)

        work3 = works[2]
        self.assertEqual(len(work3.products), 1)
        self.assertTrue("222" in work3.isbns)

    def test_agg_relworks_gtin13(self):
        agg = BookWorkAggregator(
            [
                {
                    "ISBN13": "123",
                    "GTIN_13": "123",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "GTIN_13", "IDValue": "246"}],
                        },
                    ],
                },
                {
                    "ISBN13": "246",
                    "GTIN_13": "246",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "GTIN_13", "IDValue": "246"}],
                        },
                    ],
                },
                {
                    "ISBN13": "111",
                    "GTIN_13": "111",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "GTIN_13", "IDValue": "111"}],
                        },
                        {
                            "WorkRelationCode": "something unknown",
                            "WorkIdentifiers": [{"WorkIDType": "GTIN_13", "IDValue": "246"}],
                        },
                    ],
                },
                {
                    "ISBN13": "222",
                    "GTIN_13": "222",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "GTIN_13", "IDValue": "222"}],
                        },
                    ],
                },
            ]
        )
        agg.agg_relworks()
        works = agg.get_works_from_partition(agg.uf.get_partition())
        self.assertEqual(len(works), 3)

        work1 = works[0]
        self.assertEqual(len(work1.products), 2)
        self.assertTrue("123" in work1.isbns)
        self.assertTrue("246" in work1.isbns)

        work2 = works[1]
        self.assertEqual(len(work2.products), 1)
        self.assertTrue("111" in work2.isbns)

        work3 = works[2]
        self.assertEqual(len(work3.products), 1)
        self.assertTrue("222" in work3.isbns)

    def test_agg_relworks_pid_proprietary(self):
        agg = BookWorkAggregator(
            [
                {
                    "ISBN13": "123",
                    "GTIN_13": "123",
                    "PID_Proprietary": "123",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "PID_Proprietary", "IDValue": "246"}],
                        },
                    ],
                },
                {
                    "ISBN13": "246",
                    "GTIN_13": "246",
                    "PID_Proprietary": "246",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "PID_Proprietary", "IDValue": "246"}],
                        },
                    ],
                },
                {
                    "ISBN13": "111",
                    "GTIN_13": "111",
                    "PID_Proprietary": "111",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "PID_Proprietary", "IDValue": "111"}],
                        },
                        {
                            "WorkRelationCode": "something unknown",
                            "WorkIdentifiers": [{"WorkIDType": "PID_Proprietary", "IDValue": "246"}],
                        },
                    ],
                },
                {
                    "ISBN13": "222",
                    "GTIN_13": "222",
                    "PID_Proprietary": "222",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "PID_Proprietary", "IDValue": "222"}],
                        },
                    ],
                },
            ]
        )
        agg.agg_relworks()
        works = agg.get_works_from_partition(agg.uf.get_partition())
        self.assertEqual(len(works), 3)

        work1 = works[0]
        self.assertEqual(len(work1.products), 2)
        self.assertTrue("123" in work1.isbns)
        self.assertTrue("246" in work1.isbns)

        work2 = works[1]
        self.assertEqual(len(work2.products), 1)
        self.assertTrue("111" in work2.isbns)

        work3 = works[2]
        self.assertEqual(len(work3.products), 1)
        self.assertTrue("222" in work3.isbns)

    def test_agg_products(self):
        agg = BookWorkAggregator(
            [
                {
                    "ISBN13": "123",
                    "RelatedProducts": [
                        {"ProductRelationCodes": ["Alternative format", "something random"], "ISBN13": "234"}
                    ],
                },
                {
                    "ISBN13": "234",
                    "RelatedProducts": [
                        {"ProductRelationCodes": ["Alternative format", "something random"], "ISBN13": "123"}
                    ],
                },
                {
                    "ISBN13": "456",
                    "RelatedProducts": [{"ProductRelationCodes": ["random", "something random"], "ISBN13": "123"}],
                },
                {
                    "ISBN13": "789",
                    "RelatedProducts": [
                        {"ProductRelationCodes": ["Alternative format", "something random"], "ISBN13": "123"}
                    ],
                },
            ]
        )
        agg.agg_relproducts()
        works = agg.get_works_from_partition(agg.uf.get_partition())
        self.assertEqual(len(works), 2)

        work1 = works[0]
        self.assertEqual(len(work1.isbns), 3)
        self.assertTrue("123" in work1.isbns)
        self.assertTrue("234" in work1.isbns)
        self.assertTrue("789" in work1.isbns)

        work2 = works[1]
        self.assertEqual(len(work2.isbns), 1)
        self.assertTrue("456" in work2.isbns)

    def test_agg_relprod_doi(self):
        agg = BookWorkAggregator(
            [
                {
                    "ISBN13": "123",
                    "DOI": "doi",
                    "RelatedProducts": [
                        {"ProductRelationCodes": ["Alternative format", "something random"], "DOI": "234"}
                    ],
                },
            ]
        )

        self.assertRaises(Exception, agg.agg_relproducts)

    def test_agg_relprod_gtin13(self):
        agg = BookWorkAggregator(
            [
                {
                    "ISBN13": "123",
                    "GTIN_13": "123",
                    "RelatedProducts": [
                        {"ProductRelationCodes": ["Alternative format", "something random"], "GTIN_13": "234"}
                    ],
                },
                {
                    "ISBN13": "234",
                    "GTIN_13": "234",
                    "RelatedProducts": [
                        {"ProductRelationCodes": ["Alternative format", "something random"], "GTIN_13": "123"}
                    ],
                },
                {
                    "ISBN13": "456",
                    "GTIN_13": "456",
                    "RelatedProducts": [{"ProductRelationCodes": ["random", "something random"], "GTIN_13": "123"}],
                },
                {
                    "ISBN13": "789",
                    "GTIN_13": "789",
                    "RelatedProducts": [
                        {"ProductRelationCodes": ["Alternative format", "something random"], "GTIN_13": "123"}
                    ],
                },
            ]
        )

        agg.agg_relproducts()
        works = agg.get_works_from_partition(agg.uf.get_partition())
        self.assertEqual(len(works), 2)

        work1 = works[0]
        self.assertEqual(len(work1.isbns), 3)
        self.assertTrue("123" in work1.isbns)
        self.assertTrue("234" in work1.isbns)
        self.assertTrue("789" in work1.isbns)

        work2 = works[1]
        self.assertEqual(len(work2.isbns), 1)
        self.assertTrue("456" in work2.isbns)

    def test_aggregate1(self):
        agg = BookWorkAggregator(
            [
                {
                    "ISBN13": "123",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "ISBN-13", "IDValue": "246"}],
                        },
                    ],
                },
                {
                    "ISBN13": "246",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "ISBN-13", "IDValue": "246"}],
                        },
                    ],
                },
                {
                    "ISBN13": "111",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "ISBN-13", "IDValue": "111"}],
                        },
                        {
                            "WorkRelationCode": "something unknown",
                            "WorkIdentifiers": [{"WorkIDType": "ISBN-13", "IDValue": "246"}],
                        },
                    ],
                },
                {
                    "ISBN13": "222",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "Proprietary", "IDValue": "222"}],
                        },
                    ],
                },
                {
                    "ISBN13": "1010",
                    "PID_Proprietary": "1010",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "Proprietary", "IDValue": "1011"}],
                        },
                    ],
                },
                {
                    "ISBN13": "1011",
                    "PID_Proprietary": "1011",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "Proprietary", "IDValue": "1010"}],
                        },
                    ],
                },
            ]
        )

        works = agg.aggregate()
        self.assertEqual(len(works), 4)

        work1 = works[0]
        self.assertEqual(len(work1.products), 2)
        self.assertTrue("123" in work1.isbns)
        self.assertTrue("246" in work1.isbns)

        work2 = works[1]
        self.assertEqual(len(work2.products), 1)
        self.assertTrue("111" in work2.isbns)

        work3 = works[2]
        self.assertEqual(len(work3.products), 1)
        self.assertTrue("222" in work3.isbns)

        work4 = works[3]
        self.assertEqual(len(work4.products), 2)
        self.assertTrue("1010" in work4.isbns)
        self.assertTrue("1011" in work4.isbns)

    def test_aggregate2(self):
        agg = BookWorkAggregator(
            [
                {
                    "ISBN13": "123",
                    "RelatedProducts": [
                        {"ProductRelationCodes": ["Alternative format", "something random"], "ISBN13": "234"}
                    ],
                },
                {
                    "ISBN13": "234",
                    "RelatedProducts": [
                        {"ProductRelationCodes": ["Alternative format", "something random"], "ISBN13": "123"}
                    ],
                },
                {
                    "ISBN13": "456",
                    "RelatedProducts": [{"ProductRelationCodes": ["random", "something random"], "ISBN13": "123"}],
                },
                {
                    "ISBN13": "789",
                    "RelatedProducts": [
                        {"ProductRelationCodes": ["Alternative format", "something random"], "ISBN13": "123"}
                    ],
                },
                {
                    "ISBN13": "1010",
                    "PID_Proprietary": "1010",
                    "RelatedProducts": [
                        {"ProductRelationCodes": ["Alternative format", "something random"], "PID_Proprietary": "1011"}
                    ],
                },
                {
                    "ISBN13": "1011",
                    "PID_Proprietary": "1011",
                    "RelatedProducts": [
                        {"ProductRelationCodes": ["Alternative format", "something random"], "PID_Proprietary": "1010"}
                    ],
                },
            ]
        )

        works = agg.aggregate()
        self.assertEqual(len(works), 3)

        work1 = works[0]
        self.assertEqual(len(work1.isbns), 3)
        self.assertTrue("123" in work1.isbns)
        self.assertTrue("234" in work1.isbns)
        self.assertTrue("789" in work1.isbns)

        work2 = works[1]
        self.assertEqual(len(work2.isbns), 1)
        self.assertTrue("456" in work2.isbns)

        work3 = works[2]
        self.assertEqual(len(work3.isbns), 2)
        self.assertTrue("1010" in work3.isbns)
        self.assertTrue("1011" in work3.isbns)

    def test_aggregate3(self):
        agg = BookWorkAggregator([])
        lookup_table = agg.aggregate()
        self.assertEqual(len(lookup_table), 0)

    def test_agg_works_products_composite(self):
        agg = BookWorkAggregator(
            [
                {
                    "ISBN13": "111",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [
                                {"WorkIDType": "ISBN-13", "IDValue": "112"},
                            ],
                        },
                    ],
                    "RelatedProducts": [
                        {"ProductRelationCodes": ["Alternative format"], "ISBN13": "211"},
                    ],
                },
                {
                    "ISBN13": "112",
                    "RelatedWorks": [],
                    "RelatedProducts": [],
                },
                {
                    "ISBN13": "211",
                    "RelatedWorks": [],
                    "RelatedProducts": [],
                },
            ]
        )

        works = agg.aggregate()
        self.assertEqual(len(works), 1)

    def test_get_pid_idx_missing(self):
        agg = BookWorkAggregator([])
        pid_type = "ISBN13"
        pid = "123"
        isbn13_to_index = {}
        gtin13_to_product = {}

        self.assertEqual(agg.get_pid_idx(pid_type, pid), None)

        pid_type = "GTIN_13"
        self.assertEqual(agg.get_pid_idx(pid_type, pid), None)

        pid_type = "PID_Proprietary"
        self.assertEqual(agg.get_pid_idx(pid_type, pid), None)

    def test_get_pid_idx_unknown(self):
        agg = BookWorkAggregator([])
        pid_type = "somethingunknown"
        pid = "123"
        self.assertRaises(Exception, agg.get_pid_idx, pid_type, pid)

    def test_agg_relprod_missing_record(self):
        agg = BookWorkAggregator(
            [
                {
                    "ISBN13": "123",
                    "DOI": "doi",
                    "RelatedProducts": [
                        {"ProductRelationCodes": ["Alternative format", "something random"], "ISBN13": "234"}
                    ],
                },
            ]
        )
        works = agg.aggregate()
        self.assertEqual(len(agg.errors), 1)

    def test_get_works_lookup_table(self):
        agg = BookWorkAggregator(
            [
                {
                    "ISBN13": "123",
                    "RelatedProducts": [
                        {"ProductRelationCodes": ["Alternative format", "something random"], "ISBN13": "234"}
                    ],
                },
                {
                    "ISBN13": "234",
                    "RelatedProducts": [
                        {"ProductRelationCodes": ["Alternative format", "something random"], "ISBN13": "123"}
                    ],
                },
                {
                    "ISBN13": "456",
                    "RelatedProducts": [{"ProductRelationCodes": ["random", "something random"], "ISBN13": "123"}],
                },
                {
                    "ISBN13": "789",
                    "RelatedProducts": [
                        {"ProductRelationCodes": ["Alternative format", "something random"], "ISBN13": "123"}
                    ],
                },
            ]
        )

        agg.aggregate()
        lookup_table = agg.get_works_lookup_table()
        self.assertEqual(len(lookup_table), 4)

        lookup_map = {lookup_table[i]["isbn13"]: lookup_table[i]["work_id"] for i in range(len(lookup_table))}
        self.assertEqual(lookup_map["123"], "123")
        self.assertEqual(lookup_map["234"], "123")
        self.assertEqual(lookup_map["789"], "123")
        self.assertEqual(lookup_map["456"], "456")

    def log_agg_related_product_errors(self):
        agg = BookWorkAggregator(
            [
                {
                    "ISBN13": "123",
                    "RelatedProducts": [
                        {"ProductRelationCodes": ["Alternative format", "something random"], "ISBN13": "234"}
                    ],
                },
            ]
        )

        agg.aggregate()
        self.assertEqual(
            agg.errors[0],
            "Product 123 has a related product ISBN13:234 of types ['Alternative format', 'something random'] missing a product record.",
        )

    def test_log_agg_relworks_errors(self):
        agg = BookWorkAggregator(
            [
                {
                    "ISBN13": "123",
                    "GTIN_13": "123",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "ISBN-13", "IDValue": "246"}],
                        },
                    ],
                },
            ]
        )

        agg.aggregate()
        self.assertEqual(
            agg.errors[0],
            "Product ISBN13:123 is a manifestation of ISBN13:246, which is not given as a product identifier in any ONIX product record.",
        )

    def test_log_agg_relworks_errors_miss_gtin(self):
        agg = BookWorkAggregator(
            [
                {
                    "ISBN13": "123",
                    "GTIN_13": "123",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "GTIN_13", "IDValue": "246"}],
                        },
                    ],
                },
            ]
        )

        agg.aggregate()
        self.assertEqual(
            agg.errors[0],
            "Product ISBN13:123 is a manifestation of GTIN_13:246, which is not given as a product identifier in any ONIX product record.",
        )

    def test_log_get_works_lookup_table_errors(self):
        agg = BookWorkAggregator(
            [
                {
                    "ISBN13": "123",
                    "GTIN_13": "123",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "ISBN-13", "IDValue": "246"}],
                        },
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "ISBN-13", "IDValue": "789"}],
                        },
                    ],
                },
                {
                    "ISBN13": "246",
                    "GTIN_13": "123",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "ISBN-13", "IDValue": "246"}],
                        },
                    ],
                },
                {
                    "ISBN13": "789",
                    "GTIN_13": "123",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "ISBN-13", "IDValue": "789"}],
                        },
                    ],
                },
            ]
        )

        agg.aggregate()
        table = agg.get_works_lookup_table()
        self.assertEqual(len(agg.errors), 1)
        self.assertEqual(agg.errors[0][0:54], "Warning: product 123 has multiple work ID assignments:")
        self.assertEqual(len(table), 3)

    def test_filtering_duplicate_isbns(self):
        agg = BookWorkAggregator(
            [
                {
                    "ISBN13": "123",
                    "GTIN_13": "123",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "ISBN-13", "IDValue": "246"}],
                        },
                    ],
                },
                {
                    "ISBN13": "123",
                    "GTIN_13": "999",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "ISBN-13", "IDValue": "999"}],
                        },
                    ],
                },
                {
                    "ISBN13": "246",
                    "GTIN_13": "123",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "ISBN-13", "IDValue": "246"}],
                        },
                    ],
                },
                {
                    "ISBN13": "789",
                    "GTIN_13": "123",
                    "RelatedWorks": [
                        {
                            "WorkRelationCode": "Manifestation of",
                            "WorkIdentifiers": [{"WorkIDType": "ISBN-13", "IDValue": "789"}],
                        },
                    ],
                },
            ]
        )

        agg.aggregate()
        table = agg.get_works_lookup_table()
        self.assertEqual(len(agg.errors), 1)
        self.assertEqual(agg.errors[0], "ISBN 123 has duplicate product records in the ONIX data source.")
        self.assertEqual(len(table), 3)


class TestBookWorkFamilyAggregator(unittest.TestCase):
    """
    Test the BookWorkFamilyAggregator class.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def test_ctor(self):
        works = [
            BookWork(work_id="1", work_id_type="test", products=[{"ISBN13": "123"}]),
        ]

        agg = BookWorkFamilyAggregator(works)
        self.assertEqual(len(agg.works), 1)

    def test_agg_relproducts(self):
        products1 = [
            {
                "ISBN13": "123",
                "RelatedProducts": [
                    {"ProductRelationCodes": ["Replaces", "something random"], "ISBN13": "456"},
                    {"ProductRelationCodes": ["Replaces", "something random"], "ISBN13": "789"},
                ],
            },
            {"ISBN13": "456", "RelatedProducts": []},
            {"ISBN13": "147", "RelatedProducts": []},
        ]

        products2 = [
            {
                "ISBN13": "258",
                "RelatedProducts": [
                    {
                        "ProductRelationCodes": ["Is later edition of first edition", "something random"],
                        "ISBN13": "147",
                    },
                ],
            },
        ]

        products3 = [
            {
                "ISBN13": "369",
                "RelatedProducts": [
                    {"ProductRelationCodes": ["something random"], "ISBN13": "123"},
                ],
            },
        ]

        products4 = [
            {
                "ISBN13": "789",
                "RelatedProducts": [
                    {"ProductRelationCodes": ["Replaced by", "something random"], "ISBN13": "456"},
                ],
            },
        ]

        works = [
            BookWork(work_id="1", work_id_type="test", products=products1),
            BookWork(work_id="2", work_id_type="test", products=products2),
            BookWork(work_id="3", work_id_type="test", products=products3),
            BookWork(work_id="4", work_id_type="test", products=products4),
        ]

        agg = BookWorkFamilyAggregator(works)
        wfam = agg.agg_relproducts()
        self.assertEqual(len(wfam), 2)

        self.assertEqual(wfam[0].work_family_id, "1")
        self.assertEqual(wfam[0].work_family_id_type, "test")
        self.assertEqual(len(wfam[0].works), 3)
        self.assertEqual(wfam[0].works[0].work_id, "1")
        self.assertEqual(wfam[0].works[1].work_id, "2")
        self.assertEqual(wfam[0].works[2].work_id, "4")

        self.assertEqual(wfam[1].work_family_id, "3")
        self.assertEqual(wfam[1].work_family_id_type, "test")
        self.assertEqual(len(wfam[1].works), 1)
        self.assertEqual(wfam[1].works[0].work_id, "3")

    def test_aggregate_products_missing(self):
        products1 = [
            {
                "ISBN13": "123",
                "RelatedProducts": [
                    {"ProductRelationCodes": ["Replaces", "something random"], "ISBN13": "456"},
                ],
            }
        ]

        works = [
            BookWork(work_id="1", work_id_type="test", products=products1),
        ]

        agg = BookWorkFamilyAggregator(works)
        wfam = agg.aggregate()
        self.assertEqual(len(wfam), 1)
        self.assertTrue("123" in wfam[0].works[0].isbns)

    def test_aggregate_products(self):
        products1 = [
            {
                "ISBN13": "123",
                "RelatedProducts": [
                    {"ProductRelationCodes": ["Replaces", "something random"], "ISBN13": "456"},
                    {"ProductRelationCodes": ["Replaces", "something random"], "ISBN13": "789"},
                ],
            },
            {"ISBN13": "456", "RelatedProducts": []},
            {"ISBN13": "147", "RelatedProducts": []},
        ]

        products2 = [
            {
                "ISBN13": "258",
                "RelatedProducts": [
                    {
                        "ProductRelationCodes": ["Is later edition of first edition", "something random"],
                        "ISBN13": "147",
                    },
                ],
            },
        ]

        products3 = [
            {
                "ISBN13": "369",
                "RelatedProducts": [
                    {"ProductRelationCodes": ["something random"], "ISBN13": "123"},
                ],
            },
        ]

        products4 = [
            {
                "ISBN13": "789",
                "RelatedProducts": [
                    {"ProductRelationCodes": ["Replaced by", "something random"], "ISBN13": "456"},
                ],
            },
        ]

        works = [
            BookWork(work_id="1", work_id_type="test", products=products1),
            BookWork(work_id="2", work_id_type="test", products=products2),
            BookWork(work_id="3", work_id_type="test", products=products3),
            BookWork(work_id="4", work_id_type="test", products=products4),
        ]

        agg = BookWorkFamilyAggregator(works)
        wfam = agg.aggregate()
        self.assertEqual(len(wfam), 2)

        self.assertEqual(wfam[0].work_family_id, "1")
        self.assertEqual(wfam[0].works[0].work_id, "1")
        self.assertEqual(wfam[0].works[1].work_id, "2")
        self.assertEqual(wfam[0].works[2].work_id, "4")

        self.assertEqual(wfam[1].work_family_id, "3")
        self.assertEqual(wfam[1].works[0].work_id, "3")

    def test_aggregate_products_gtin13(self):
        products1 = [
            {
                "ISBN13": "123",
                "GTIN_13": "123",
                "RelatedProducts": [
                    {"ProductRelationCodes": ["Replaces", "something random"], "GTIN_13": "456"},
                    {"ProductRelationCodes": ["Replaces", "something random"], "GTIN_13": "789"},
                ],
            },
            {"ISBN13": "456", "GTIN_13": "456", "RelatedProducts": []},
            {"ISBN13": "147", "GTIN_13": "147", "RelatedProducts": []},
        ]

        products2 = [
            {
                "ISBN13": "258",
                "GTIN_13": "258",
                "RelatedProducts": [
                    {
                        "ProductRelationCodes": ["Is later edition of first edition", "something random"],
                        "GTIN_13": "147",
                    },
                ],
            },
        ]

        products3 = [
            {
                "ISBN13": "369",
                "GTIN_13": "369",
                "RelatedProducts": [
                    {"ProductRelationCodes": ["something random"], "GTIN_13": "123"},
                ],
            },
        ]

        products4 = [
            {
                "ISBN13": "789",
                "GTIN_13": "789",
                "RelatedProducts": [
                    {"ProductRelationCodes": ["Replaced by", "something random"], "GTIN_13": "456"},
                ],
            },
        ]

        works = [
            BookWork(work_id="1", work_id_type="test", products=products1),
            BookWork(work_id="2", work_id_type="test", products=products2),
            BookWork(work_id="3", work_id_type="test", products=products3),
            BookWork(work_id="4", work_id_type="test", products=products4),
        ]

        agg = BookWorkFamilyAggregator(works)
        wfam = agg.aggregate()
        self.assertEqual(len(wfam), 2)

        self.assertEqual(wfam[0].work_family_id, "1")
        self.assertEqual(wfam[0].works[0].work_id, "1")
        self.assertEqual(wfam[0].works[1].work_id, "2")
        self.assertEqual(wfam[0].works[2].work_id, "4")

        self.assertEqual(wfam[1].work_family_id, "3")
        self.assertEqual(wfam[1].works[0].work_id, "3")

    def test_aggregate_products_pid_proprietary(self):
        products1 = [
            {
                "ISBN13": "123",
                "GTIN_13": "123",
                "PID_Proprietary": "123",
                "RelatedProducts": [
                    {"ProductRelationCodes": ["Replaces", "something random"], "PID_Proprietary": "456"},
                    {"ProductRelationCodes": ["Replaces", "something random"], "PID_Proprietary": "789"},
                ],
            },
            {"ISBN13": "456", "GTIN_13": "456", "PID_Proprietary": "456", "RelatedProducts": []},
            {"ISBN13": "147", "GTIN_13": "147", "PID_Proprietary": "147", "RelatedProducts": []},
        ]

        products2 = [
            {
                "ISBN13": "258",
                "GTIN_13": "258",
                "PID_Proprietary": "258",
                "RelatedProducts": [
                    {
                        "ProductRelationCodes": ["Is later edition of first edition", "something random"],
                        "PID_Proprietary": "147",
                    },
                ],
            },
        ]

        products3 = [
            {
                "ISBN13": "369",
                "GTIN_13": "369",
                "PID_Proprietary": "369",
                "RelatedProducts": [
                    {"ProductRelationCodes": ["something random"], "PID_Proprietary": "123"},
                ],
            },
        ]

        products4 = [
            {
                "ISBN13": "789",
                "GTIN_13": "789",
                "PID_Proprietary": "789",
                "RelatedProducts": [
                    {"ProductRelationCodes": ["Replaced by", "something random"], "PID_Proprietary": "456"},
                ],
            },
        ]

        works = [
            BookWork(work_id="1", work_id_type="test", products=products1),
            BookWork(work_id="2", work_id_type="test", products=products2),
            BookWork(work_id="3", work_id_type="test", products=products3),
            BookWork(work_id="4", work_id_type="test", products=products4),
        ]

        agg = BookWorkFamilyAggregator(works)
        wfam = agg.aggregate()
        self.assertEqual(len(wfam), 2)

        self.assertEqual(wfam[0].work_family_id, "1")
        self.assertEqual(wfam[0].works[0].work_id, "1")
        self.assertEqual(wfam[0].works[1].work_id, "2")
        self.assertEqual(wfam[0].works[2].work_id, "4")

        self.assertEqual(wfam[1].work_family_id, "3")
        self.assertEqual(wfam[1].works[0].work_id, "3")

    def test_get_wid_unsupported(self):
        products1 = [
            {
                "ISBN13": "123",
                "xid": "123",
                "RelatedProducts": [
                    {"ProductRelationCodes": ["Replaces", "something random"], "xid": "456"},
                ],
            }
        ]

        works = [BookWork(work_id="1", work_id_type="test", products=products1)]
        agg = BookWorkFamilyAggregator(works)
        self.assertRaises(Exception, agg.aggregate)

    def test_get_wid_idx_unsupported(self):
        works = [BookWork(work_id="1", work_id_type="test", products=[])]
        agg = BookWorkFamilyAggregator(works)
        self.assertRaises(Exception, agg.get_wid_idx, "unknown", "123", {}, {}, {})

    def test_get_wid_idx_missing_isbn(self):
        works = [BookWork(work_id="1", work_id_type="test", products=[])]
        agg = BookWorkFamilyAggregator(works)
        pid_type = "ISBN13"
        pid = "123"
        isbn_table = {}
        gtin_table = {}
        proprietary_table = {}
        self.assertEqual(agg.get_wid_idx(pid_type, pid, isbn_table, gtin_table, proprietary_table), None)

    def test_get_wid_idx_missing_gtin(self):
        works = [BookWork(work_id="1", work_id_type="test", products=[])]
        agg = BookWorkFamilyAggregator(works)
        pid_type = "GTIN_13"
        pid = "123"
        isbn_table = {}
        gtin_table = {}
        proprietary_table = {}
        self.assertEqual(agg.get_wid_idx(pid_type, pid, isbn_table, gtin_table, proprietary_table), None)

    def test_get_wid_idx_missing_pid_proprietary(self):
        works = [BookWork(work_id="1", work_id_type="test", products=[])]
        agg = BookWorkFamilyAggregator(works)
        pid_type = "PID_Proprietary"
        pid = "123"
        isbn_table = {}
        gtin_table = {}
        proprietary_table = {}
        self.assertEqual(agg.get_wid_idx(pid_type, pid, isbn_table, gtin_table, proprietary_table), None)

    def test_get_works_family_lookup_table(self):
        products1 = [
            {
                "ISBN13": "123",
                "RelatedProducts": [
                    {"ProductRelationCodes": ["Replaces", "something random"], "ISBN13": "456"},
                    {"ProductRelationCodes": ["Replaces", "something random"], "ISBN13": "789"},
                ],
            },
            {"ISBN13": "456", "RelatedProducts": []},
            {"ISBN13": "147", "RelatedProducts": []},
        ]

        products2 = [
            {
                "ISBN13": "258",
                "RelatedProducts": [
                    {
                        "ProductRelationCodes": ["Is later edition of first edition", "something random"],
                        "ISBN13": "147",
                    },
                ],
            },
        ]

        products3 = [
            {
                "ISBN13": "369",
                "RelatedProducts": [
                    {"ProductRelationCodes": ["something random"], "ISBN13": "123"},
                ],
            },
        ]

        products4 = [
            {
                "ISBN13": "789",
                "RelatedProducts": [
                    {"ProductRelationCodes": ["Replaced by", "something random"], "ISBN13": "456"},
                ],
            },
        ]

        works = [
            BookWork(work_id="1", work_id_type="test", products=products1),
            BookWork(work_id="2", work_id_type="test", products=products2),
            BookWork(work_id="3", work_id_type="test", products=products3),
            BookWork(work_id="4", work_id_type="test", products=products4),
        ]

        agg = BookWorkFamilyAggregator(works)
        agg.aggregate()
        lookup_table = agg.get_works_family_lookup_table()
        self.assertEqual(len(lookup_table), 6)

        lookup_map = {lookup_table[i]["isbn13"]: lookup_table[i]["work_family_id"] for i in range(len(lookup_table))}

        self.assertEqual(lookup_map["123"], "1")
        self.assertEqual(lookup_map["456"], "1")
        self.assertEqual(lookup_map["147"], "1")
        self.assertEqual(lookup_map["258"], "1")
        self.assertEqual(lookup_map["789"], "1")
        self.assertEqual(lookup_map["369"], "3")
